__all__ = ('Compound', 'CompoundMetaType', 'Theory')

import reprlib
from itertools import chain
from types import GetSetDescriptorType, MemberDescriptorType, FunctionType, BuiltinMethodType, MethodType

from .analyzer import CallableAnalyzer


SLOT_TYPES = frozenset((GetSetDescriptorType, MemberDescriptorType))

COMPOUND_META_TYPE_MAP = {}


def is_slot_of(attribute_name, attribute_value, type_):
    """
    Returns whether `value` is a slot of any class from `types`.
    
    Parameters
    ----------
    attribute_name : `str`
        The attribute's name.
    attribute_value : `any`
        Value to be checked to be a slot of any type from `types`.
    type_ : `type_`
        The parent type.
    
    Returns
    -------
    is_slot_of : `bool`
    """
    for slot_type in SLOT_TYPES:
        if isinstance(attribute_value, slot_type):
            break
    else:
        return False
    
    slot_owner_type = getattr(attribute_value, '__objclass__', None)
    if slot_owner_type is None:
        return False
    
    name = getattr(attribute_value, '__name__', None)
    if name is None:
        return False
    
    if name != attribute_name:
        return False
    
    if slot_owner_type not in type_.__mro__:
        return False
    
    return True


def _redirect_constructor(cls, type_name, base_types, type_attributes):
    """
    Used by ``ignore_meta_type`` to redirect `metatype.__new__` calls.
    
    Parameters
    ----------
    cls : `type`
        The base type.
    type_name : `str`
        The created type's future name.
    base_types : `tuple` of `type`
        Base types to inherit from.
    type_attributes : `dict` of (`str`, `Any`) items
        Attributes of the created type.
    
    Returns
    -------
    created_type : `type`
        The created type.
    """
    return type.__new__(cls.__wrapped_meta_type__, type_name, base_types, type_attributes)


def ignore_meta_type(meta_type):
    """
    Ignores meta type on creation.
    
    Parameters
    ----------
    meta_type : `type`
        The meta type to ignore.
    
    Returns
    -------
    redirect : `type`
    """
    return type(
        f'{meta_type}Redirect',
        (
            meta_type,
        ), {
            '__new__': _redirect_constructor,
            '__wrapped_meta_type__': meta_type,
        }
    )


class Theory:
    """
    Represents a theoretical value, which should be implemented by other types.
    
    Attributes
    ----------
    theorem : `any`
        The legend itself.
    """
    __slots__ = ('theorem',)
    
    def __new__(cls, theorem):
        self = object.__new__(cls)
        self.theorem = theorem
        return self
    
    
    def __call__(self, *positional_parameters, **keyword_parameters):
        """
        Calls the theory.
        
        Parameters
        ----------
        *positional_parameters : Positional parameters
            Positional parameters to call the theory with.
        *keyword_parameters : Keyword parameters
            Keyword parameters to call the theory with. 
        
        Returns
        -------
        return_value : `any`
            The returned value by the theory.
        
        Raises
        ------
        BaseException
            Exception raised by the theory.
        """
        return self.theorem(*positional_parameters, **keyword_parameters)
    
    
    def __dir__(self):
        """Returns the attributes of the theory including the theorem's."""
        return sorted({*object.__dir__(self), *dir(self.theorem)})
    
    
    def __getattr__(self, attribute_name):
        """Returns the attribute of the theorem."""
        return getattr(self.theorem, attribute_name)


def collect_attributes_per_type_from_type_attributes(type_attributes):
    """
    Collects attributes per type from type attribute dictionary.
    
    Parameters
    ----------
    type_attributes : `dict` of (`str`, `Any`) items
        Type attributes of a class.
    
    Returns
    -------
    instance_attributes : `None` or `set` of `name`
        A set of attribute names mentioned in the type attributes.
    theoretical_type_attributes : `None`, `dict` of (`str`, `Any`) items
        Theoretical defined type attributes.
    implemented_type_attributes : `None`, `dict` of (`str`, `Any`) items
        Implemented type attributes.
    leftover_type_attributes : `dict` of (`str`, `Any`) items
        Additional type attributes which should not be touched.
    """
    instance_attributes = type_attributes.pop('__slots__', None)
    if instance_attributes is None:
        instance_attributes = None
    
    # Can happen that you mess up, but you should not focus on this error, go go maniac!
    elif isinstance(instance_attributes, str):
        instance_attributes = {instance_attributes}
    
    else:
        instance_attributes = {*instance_attributes}
        if not instance_attributes:
            instance_attributes = None
    
    
    theoretical_type_attributes = None
    implemented_type_attributes = None
    leftover_type_attributes = {}
    
    
    for attribute_name, attribute_value in type_attributes.items():
        if attribute_name == '__annotations__':
            continue
        
        if attribute_name in {'__module__', '__qualname__'}:
            leftover_type_attributes[attribute_name] = attribute_value
            continue
        
        if (instance_attributes is not None) and (attribute_name in instance_attributes):
            continue
        
        if isinstance(attribute_value, Theory):
            if (theoretical_type_attributes is None):
                theoretical_type_attributes = {}
            
            theoretical_type_attributes[attribute_name] = attribute_value.theorem
        
        else:
            if (implemented_type_attributes is None):
                implemented_type_attributes = {}
            
            implemented_type_attributes[attribute_name] = attribute_value
    
    return instance_attributes, theoretical_type_attributes, implemented_type_attributes, leftover_type_attributes


class CompoundMetaType(type):
    """
    Stops class creation and redirects it into compound component creation instead.
    """
    __source_meta_type__ = None
    
    def __new__(cls, type_name, base_types, type_attributes, *, build=False, **keyword_parameters):
        """
        Creates a new compound component.
        
        Parameters
        ----------
        type_name : `str`
            The created component's name.
        base_types : `tuple` of `type`
            types from which the type inherits from.
        type_attributes : `dict` of (`str`, `Any`) items
            Attributes defined in the type body.
        build : `bool` = `False`, Optional (Keyword only)
            Whether the type should be built.
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Returns
        -------
        compound : ``CompoundMetaType``
            The created compound.
        
        Raises
        ------
        TypeError
            - If a base-type is a not a type.
            - Unexpected base-type.
            - If a base-type is a not a type.
            - Only the first type can bo a non-component type.
            - Get-descriptor mismatch.
            - Instance attribute overwritten with other get-property.
            - Both method is implemented.
            - Method parameter mismatch.
            - If a type  attribute is never implemented.
            - Constant value mismatch.
            - Keyword parameters are not supported when building a component.
        """
        if build:
            compound = cls.__build__(type_name, base_types, type_attributes, keyword_parameters)
        
        else:
            if keyword_parameters:
                raise TypeError(
                    f'Keyword parameters are not supported when building a component. '
                    f'Got type_name={type_name!r}; keyword_parameters={keyword_parameters!r}.'
                )
            
            compound = cls.__create_component__(type_name, base_types, type_attributes)
        
        return compound
    
    
    @classmethod
    def __build__(cls, type_name, base_types, type_attributes, keyword_parameters):
        """
        Merges the compounds together.
        
        Parameters
        ----------
        type_name : `str`
            The created type's future name.
        base_types : `tuple` of `type`
            The types to inherit from.
        type_attributes : `dict` of (`str`, `Any`) items
            The attributes defined inside of the type body.
        keyword_parameters : `dict` of (`str`, `Any`) items
            Additional keyword parameters.
        
        Returns
        -------
        compound : `type`
            The created compound.
        
        Raises
        ------
        TypeError
            - If a base-type is a not a type.
            - Only the first type can bo a non-component type.
            - Get-descriptor mismatch.
            - Instance attribute overwritten with other get-property.
            - Both method is implemented.
            - Method parameter mismatch.
            - If a type  attribute is never implemented.
            - Constant value mismatch.
        """
        source_meta_type = cls.__source_meta_type__
        if (source_meta_type is None):
            build_type = type
            created_type = cls
        
        else:
            build_type = source_meta_type
            created_type = source_meta_type
        
        
        new_base_types, new_type_attributes = build_type_structure(type_name, base_types, type_attributes)
        
        return build_type.__new__(created_type, type_name, new_base_types, new_type_attributes, **keyword_parameters)
    
    
    @classmethod
    def __create_component__(cls, type_name, base_types, type_attributes):
        """
        Creates a new compound component.
        
        Parameters
        ----------
        type_name : `str`
            The created component's name.
        base_types : `tuple` of `type`
            types from which the type inherits from.
        type_attributes : `dict` of (`str`, `Any`) items
            Attributes defined in the type body.
        
        Returns
        -------
        compound : ``CompoundMetaType``
            The created compound.
        
        Raises
        ------
        TypeError
            - If a base-type is a not a type.
            - Unexpected base-type.
        """
        for base_type in base_types:
            if not isinstance(base_type, type):
                raise TypeError(
                    f'Base types can only be `type` instances, got {base_type!r}; type_name={type_name!r};'
                    f'base_types={base_types}.'
                )
            
            if base_type is object:
                continue
            
            if base_type is Compound:
                continue
            
            raise TypeError(
                f'Base types can only be `object` or `{Compound.__name__}`, got {base_type!r};'
                f'type_name={type_name!r}; base_types={base_types}.'
            )
        
        
        (
            instance_attributes,
            theoretical_type_attributes,
            implemented_type_attributes,
            leftover_type_attributes,
        ) = collect_attributes_per_type_from_type_attributes( 
            type_attributes,
        )
        
        return type.__new__(
            cls,
            type_name,
            (
                Compound,
            ),
            {
                **leftover_type_attributes,
                '__instance_attributes__': instance_attributes,
                '__theoretical_type_attributes__': theoretical_type_attributes,
                '__implemented_type_attributes__': implemented_type_attributes,
            }
        )
    
    
    @classmethod
    def with_(cls, other):
        """
        Meta type merging to allow compounding
        
        Parameters
        ----------
        other : `type`
            The other type to merge with.
        
        Raises
        ------
        TypeError
            - If `other` is not a type.
        """
        if not isinstance(other, type):
            raise TypeError(
                f'`other` must be a type, got {other!r}.'
            )
        
        if not issubclass(other, type):
            other = type(other)
        
        if other is type:
            return cls
        
        if hasattr(other, '__source_meta_type__'):
            return other
        
        try:
            return COMPOUND_META_TYPE_MAP[other]
        except KeyError:
            pass
        
        merge_type = type(
            f'{other.__name__}CompoundMergeType',
            (
                cls, other,
            ), {
                '__source_meta_type__': other,
            }
        )
        COMPOUND_META_TYPE_MAP[other] = merge_type
        return merge_type
    
    
    def __dir__(sub_type):
        """Lists the attributes of the compound component."""
        directory = {*type.__dir__(sub_type)}
        
        type_instance_attributes = sub_type.__instance_attributes__
        if (type_instance_attributes is not None):
            directory.update(type_instance_attributes)
        
        theoretical_type_attributes = sub_type.__theoretical_type_attributes__
        if (theoretical_type_attributes is not None):
            directory.update(theoretical_type_attributes.keys())
        
        implemented_type_attributes = sub_type.__implemented_type_attributes__
        if (implemented_type_attributes is not None):
            directory.update(implemented_type_attributes.keys())
        
        return sorted(directory)
    
    
    def __getattr__(sub_type, attribute_name):
        """Gets an attribute of the sub type. It can be either instance attribute, implemented or theoretical."""
        type_instance_attributes = sub_type.__instance_attributes__
        if (type_instance_attributes is not None):
            if attribute_name in type_instance_attributes:
                return NotImplemented
        
        theoretical_type_attributes = sub_type.__theoretical_type_attributes__
        if (theoretical_type_attributes is not None):
            try:
                return theoretical_type_attributes[attribute_name]
            except KeyError:
                pass
        
        implemented_type_attributes = sub_type.__implemented_type_attributes__
        if (implemented_type_attributes is not None):
            try:
                return implemented_type_attributes[attribute_name]
            except KeyError:
                pass
        
        raise AttributeError(attribute_name)


class Compound(metaclass=ignore_meta_type(CompoundMetaType)):
    """
    Compound component base type. Inherit from it to create a type component.
    """
    __slots__ = ()
    
    __instance_attributes__ = None
    __theoretical_type_attributes__ = None
    __implemented_type_attributes__ = None
    
    def __new__(cls):
        """
        Compound components do not support instancing.
        
        Sub-typing is supported, but the attributes are only proxied. Sub-types have no meaning by themselves.
        
        Raises
        -------
        RuntimeError
        """
        raise RuntimeError(
            f'{cls.__name__} does not supports instancing.'
        )


class CompoundLayer:
    """
    Represents a layer when building a type.
    
    Attributes
    ----------
    implemented_type_attributes : `None`, `dict` of (`str`, `Any`) items
        Implemented type attributes by the component.
    instance_attributes : `None`, `set` of `str`
        The instance attributes to implement.
    implemented : `bool`
        Whether the layer is already implemented.
    theoretical_type_attributes : `None`, `dict` of (`str`, `Any`) items
        Theoretical type attributes which an other component should implement.
    type_name : `str`
        The represented type's or the component's name.
    """
    __slots__ = (
        'implemented_type_attributes', 'instance_attributes', 'implemented', 'theoretical_type_attributes',
        'type_name'
    )
    
    def __new__(
        cls, type_name, instance_attributes, theoretical_type_attributes, implemented_type_attributes,
        implemented
    ):
        """
        Parameters
        ----------
        type_name : `str`
            The represented type's or the component's name.
        instance_attributes : `None`, `set` of `str`
            The instance attributes to implement.
        theoretical_type_attributes : `None`, `dict` of (`str`, `Any`) items
            Theoretical type attributes which an other component should implement.
        implemented_type_attributes : `None`, `dict` of (`str`, `Any`) items
            Implemented type attributes by the component.
        implemented : `bool`
            Whether the layer is already implemented.
        """
        self = object.__new__(cls)
        self.type_name = type_name
        self.instance_attributes = instance_attributes
        self.theoretical_type_attributes = theoretical_type_attributes
        self.implemented_type_attributes = implemented_type_attributes
        self.implemented = implemented
        return self
    
    
    @classmethod
    def from_component(cls, component, *, implemented=False):
        """
        Creates a new component layer from the given component.
        
        Parameters
        ----------
        component : ``CompoundMetaType``
            Compound component.
        implemented : `bool` = `False`, Optional (Keyword only)
            Whether the layer is already implemented.
        
        Returns
        -------
        self : ``ComponentLayer``
        """
        instance_attributes = component.__instance_attributes__
        if (instance_attributes is not None):
            instance_attributes = instance_attributes.copy()
        
        theoretical_type_attributes = component.__theoretical_type_attributes__
        if (theoretical_type_attributes is not None):
            theoretical_type_attributes = theoretical_type_attributes.copy()
        
        implemented_type_attributes = component.__implemented_type_attributes__
        if (implemented_type_attributes is not None):
            implemented_type_attributes = implemented_type_attributes.copy()
        
        return cls(
            component.__name__,
            instance_attributes,
            theoretical_type_attributes,
            implemented_type_attributes,
            implemented,
        )
    
    
    @classmethod
    def from_type(cls, type_, *, implemented=False):
        """
        Creates a new component layer from the given type.
        
        Parameters
        ----------
        type_ : `type`
            Base type to inherit from.
        implemented : `bool` = `False`, Optional (Keyword only)
            Whether the layer is already implemented.
        
        Returns
        -------
        self : ``ComponentLayer``
        """
        instance_attributes = None
        type_attributes = None
        
        type_directory = type_.__dict__
        for attribute_name in dir(type_):
            if attribute_name in {'__doc__', '__slots__', '__annotations__', '__module__', '__qualname__', '__class__'}:
                continue
            
            try:
                attribute_value = type_directory[attribute_name]
            except KeyError:
                attribute_value = getattr(type_, attribute_name)
            
            try:
                default_implementation =  getattr(object, attribute_name)
            except AttributeError:
                pass
            else:
                if attribute_value is default_implementation:
                    continue
            
            if attribute_name == '__new__':
                if type(attribute_value) is staticmethod:
                    attribute_value = attribute_value.__func__
            
            elif attribute_name in {'__subclasshook__', '__init_subclass__'}:
                if type(attribute_value) is MethodType:
                    attribute_value = attribute_value.__func__
                    if attribute_value is getattr(getattr(object, attribute_name), '__func__', None):
                        continue
                    
                elif type(attribute_value) is BuiltinMethodType:
                    continue
            
            
            if is_slot_of(attribute_name, attribute_value, type_):
                if instance_attributes is None:
                    instance_attributes = set()
                    
                instance_attributes.add(attribute_name)
            
            else:
                if type_attributes is None:
                    type_attributes = {}
                    
                type_attributes[attribute_name] = attribute_value
        
        
        return cls(
            type_.__name__,
            instance_attributes,
            None,
            type_attributes,
            implemented
        )
    
    
    @classmethod
    def from_name_and_attributes(cls, type_name, type_attributes,*, implemented=False):
        """
        Creates a new component layer from the given `type_name` - `type_attributes` pair.
        
        Parameters
        ----------
        type_name : `str`
            The created component's name.
        type_attributes : `dict` of (`str`, `Any`) items
            Attributes defined in the type body.
        implemented : `bool` = `False`, Optional (Keyword only)
            Whether the layer is already implemented.
        
        Returns
        -------
        self : ``ComponentLayer``
            The created component layer.
        leftover_type_attributes : `dict` of (`str`, `Any`) items
            Additional type attributes which should not be touched.
        """
        (
            instance_attributes,
            theoretical_type_attributes,
            implemented_type_attributes,
            leftover_type_attributes,
        ) = collect_attributes_per_type_from_type_attributes(type_attributes)
        
        
        return (
            cls(
                type_name,
                instance_attributes,
                theoretical_type_attributes,
                implemented_type_attributes,
                implemented,
            ),
            leftover_type_attributes,
        )
    
    
    def iter_attribute_names(self):
        """
        Iterates over the attributes mentioned in the layer.
        
        This method is an iterable generator.
        
        Yields
        ------
        attribute_name : `str`
        """
        instance_attributes = self.instance_attributes
        if (instance_attributes is not None):
            yield from instance_attributes
        
        theoretical_type_attributes = self.theoretical_type_attributes
        if (theoretical_type_attributes is not None):
            yield from theoretical_type_attributes.keys()
        
        implemented_type_attributes = self.implemented_type_attributes
        if (implemented_type_attributes is not None):
            yield from implemented_type_attributes
    
    
    def get_implementation_detail_of(self, attribute_name, *, exhaust=False):
        """
        Gets the implementation details for the given attribute name.
        
        Parameters
        ----------
        attribute_name : `str`
            The attribute's name to get it's implementation of.
        exhaust : `bool` = `False`, Optional (Keyword only)
            Whether the attribute should be discarded up use.
        
        Returns
        -------
        implementation_detail : `None`, ``ImplementationDetail``
        """
        instance_attributes = self.instance_attributes
        if (instance_attributes is not None) and (attribute_name in instance_attributes):
            if exhaust:
                instance_attributes.discard(attribute_name)
                if not instance_attributes:
                    self.instance_attributes = None
            
            return ImplementationDetail(True, self.implemented, None, self.type_name, attribute_name)
        
        
        theoretical_type_attributes = self.theoretical_type_attributes
        if (theoretical_type_attributes is not None):
            try:
                implementation = theoretical_type_attributes[attribute_name]
            except KeyError:
                pass
            else:
                if exhaust:
                    theoretical_type_attributes.pop(attribute_name)
                    if not theoretical_type_attributes:
                        self.theoretical_type_attributes = None
                
                return ImplementationDetail(False, False, implementation, self.type_name, attribute_name)
        
        
        implemented_type_attributes = self.implemented_type_attributes
        if (implemented_type_attributes is not None):
            try:
                implementation = implemented_type_attributes[attribute_name]
            except KeyError:
                pass
            else:
                if exhaust:
                    implemented_type_attributes.pop(attribute_name)
                    if not implemented_type_attributes:
                        self.implemented_type_attributes = None
                
                return ImplementationDetail(False, True, implementation, self.type_name, attribute_name)
        
        
        return None


class ImplementationDetail:
    """
    Represents details about and attribute's implementation.
    
    Attributes
    ----------
    instance_attribute : `bool`
        Whether the implementation is an instance attribute.
    implemented : `bool`
        Whether the instance attribute is implemented.
    implementation : `None`, `Any`
        The implementation.
    implementation_source : `None`, `str`
        From where the implementation is coming from.
    name : `str`
        The name of the represented attribute.
    """
    __slots__ = ('instance_attribute', 'implemented', 'implementation', 'implementation_source', 'name')
    
    def __new__(cls, instance_attribute, implemented, implementation, implementation_source, name):
        """
        Creates a new ``ImplementationDetail``.
        
        Parameters
        ----------
        instance_attribute : `bool`
            Whether the implementation is an instance attribute.
        implemented : `bool`
            Whether the instance attribute is implemented.
        implementation : `None`, `Any`
            The implementation.
        implementation_source : `str`
            From where the implementation is coming from.
        name : `str`
            The name of the represented attribute.
        """
        self = object.__new__(cls)
        self.instance_attribute = instance_attribute
        self.implemented = implemented
        self.implementation = implementation
        self.implementation_source = implementation_source
        self.name = name
        return self
    
    
    def __repr__(self):
        """Returns the implementation detail's representation."""
        repr_parts = ['<', self.__class__.__name__, ' ', self.implementation_source, '.', self.name, ' ']
        
        if self.instance_attribute:
            repr_parts.append('instance_attribute')
        
        elif self.is_get_descriptor():
            repr_parts.append('get_descriptor')
        
        elif self.is_method():
            repr_parts.append('method')
            
        
        
        if self.implemented:
            repr_parts.append(', implemented=True')
            
            repr_parts.append(', implementation=')
            
            implementation = self.implementation
            if isinstance(implementation, FunctionType):
                implementation_representation = repr(implementation)
            else:
                implementation_representation = reprlib.repr(implementation)
            repr_parts.append(implementation_representation)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def is_get_descriptor(self):
        """
        Returns whether the implementation detail is a property.
        
        Returns
        -------
        is_get_descriptor : `bool`
        """
        if self.instance_attribute:
            return True
        
        implementation = self.implementation
        if (implementation is None):
            return False
        
        return isinstance(implementation, property)
    
    
    def is_method(self):
        """
        Returns whether the implementation is a method.
        
        Returns
        -------
        is_method : `bool`
        """
        if self.instance_attribute:
            return False
        
        implementation = self.implementation
        if (implementation is None):
            return False
        
        return isinstance(implementation, FunctionType)
    
    
    def is_constant(self):
        """
        Returns whether the implementation is a constant.
        
        Returns
        -------
        is_method : `bool`
        """
        if self.instance_attribute:
            return False
        
        implementation = self.implementation
        if implementation is None:
            return True
        
        if implementation is NotImplemented:
            return False
        
        if isinstance(implementation, FunctionType):
            return False
        
        if (
            hasattr(implementation, '__get__') or
            hasattr(implementation, '__set__') or
            hasattr(implementation, '__delete__')
        ):
            return False
        
        return True
    
    
    def is_implemented(self):
        """
        Returns whether the implementation detail was ever really implemented.
        
        Returns
        -------
        is_implemented : `bool`
        """
        if self.instance_attribute:
            return True
        
        if not self.implemented:
            return False
        
        if self.implementation is NotImplemented:
            return False
        
        return True
    
    
    def apply_layer_implementation(self, layer_implementation, attribute_name):
        """
        Applies an implementation of an other layer.
        
        Parameters
        ----------
        layer_implementation : ``ImplementationDetail``
            The implementation detail to apply.
        attribute_name : `str`
            The respective attribute's name.
        
        Raises
        ------
        TypeError
            - Get-descriptor mismatch.
            - Instance attribute overwritten with other get-property.
            - Constant value mismatch.
        """
        _check_get_descriptor_match(self, layer_implementation)
        _check_instance_attribute_overwritability(self, layer_implementation)
        _check_method_match(self, layer_implementation)
        _check_constant_match(self, layer_implementation)
        _maybe_transfer_method_implementation(self, layer_implementation)
        _maybe_transfer_implementation(self, layer_implementation)


class CompoundLogicUnit:
    """
    Defines lookup logic unit for building component types.
    
    The existence of these units only comes from design, since some nice OOP design makes the implementation simpler.
    
    Attributes
    ----------
    attribute_names : `set` of `str`
        The attribute names of the layer to resolve.
    base_layer : ⅞none`, ``ComponentLayer``
        The base type's layer.
    component_layers : `list` of ``ComponentLayer``
        Component layers.
    type_layer : ``ComponentLayer``
        The first priority layer.
    """
    __slots__ = ('attribute_names', 'base_layer', 'component_layers', 'type_layer')
    
    def __new__(cls, type_layer):
        """
        Creates a new compound logic unit.
        
        Parameters
        ----------
        type_layer : ``ComponentLayer``
            The first priority layer.
        """
        attribute_names = {*type_layer.iter_attribute_names()}
        
        self = object.__new__(cls)
        self.attribute_names = attribute_names
        self.base_layer = None
        self.component_layers = []
        self.type_layer = type_layer
        
        return self
    
    
    def add_component_layers(self, layer):
        """
        Adds a component layer to the logic.
        
        > New layers are always added to the end, so order matters.
        
        Parameters
        ----------
        layer : ``ComponentLayer``
        """
        self.component_layers.append(layer)
        self.attribute_names.update(layer.iter_attribute_names())
    
    
    def add_base_layer(self, layer):
        """
        Adds a base type layer to the type.
        
        Parameters
        ----------
        layer : ``ComponentLayer``
        """
        self.base_layer = layer
        self.attribute_names.update(layer.iter_attribute_names())
    
    
    def iter_all_layer(self):
        """
        Iterates over all the registered layers in order.
        
        This method is an iterable generator.
        
        Yields
        ------
        layer : ``CompoundLayer``
        """
        yield self.type_layer
        
        base_layer = self.base_layer
        if (base_layer is not None):
            yield base_layer
        
        yield from self.component_layers
    
    
    def exhaust_attribute_names(self):
        """
        Exhaust the attribute names of the compound logic.
        
        This method is an iterable generator.
        
        Yields
        ------
        attribute_name : `str`
        """
        attribute_names = self.attribute_names
        while attribute_names:
            yield attribute_names.pop()
    
    
    def try_get_base_implementation_for(self, attribute_name):
        """
        tries to get implementation details for the base implementation of the given `attribute_name`.
        
        Returns
        -------
        implementation_detail : `None, ``ImplementationDetail``
        
        Raises
        ------
        TypeError
            - Get-descriptor mismatch.
            - Instance attribute overwritten with other get-property.
            - Constant value mismatch.
        """
        type_layer_implementation_detail = self.type_layer.get_implementation_detail_of(attribute_name, exhaust=True)
        
        base_layer = self.base_layer
        if (base_layer is None):
            base_layer_implementation_detail = None
        
        else:
            base_layer_implementation_detail = base_layer.get_implementation_detail_of(attribute_name, exhaust=True)
        
        
        if (type_layer_implementation_detail is None):
            if (base_layer_implementation_detail is None):
                return None
            
            else:
                return base_layer_implementation_detail
        
        else:
            if (base_layer_implementation_detail is None):
                return type_layer_implementation_detail
        
        
        _check_get_descriptor_match(base_layer_implementation_detail, type_layer_implementation_detail)
        _check_instance_attribute_overwritability(base_layer_implementation_detail, type_layer_implementation_detail)
        _check_constant_match(base_layer_implementation_detail, type_layer_implementation_detail)
        
        _maybe_transfer_instance_attribute_implementation(
            base_layer_implementation_detail, type_layer_implementation_detail
        )
        
        _maybe_transfer_implementation(base_layer_implementation_detail, type_layer_implementation_detail)
        
        return type_layer_implementation_detail


def _check_get_descriptor_match(source_implementation_detail, added_implementation_detail):
    """
    Checks whether the source one is a get descriptor but the other one is not.
    
    Parameters
    ----------
    source_implementation_detail : ``ImplementationDetail``
        Priority source implementation.
    added_implementation_detail : ``ImplementationDetail``
        The added implementation.
    
    Raises
    ------
    TypeError
        - Get-descriptor mismatch.
    """
    if source_implementation_detail.is_get_descriptor() != added_implementation_detail.is_get_descriptor():
        raise TypeError(
            f'Implementation of `{source_implementation_detail.name}` at'
            f'`{source_implementation_detail.implementation_source}` is defined as a `get-descriptor` '
            f', meanwhile at `{added_implementation_detail.implementation_source}` is not. '
            f'Got `{source_implementation_detail!r}`; `{added_implementation_detail!r}`.'
        )


def _check_instance_attribute_overwritability(source_implementation_detail, added_implementation_detail):
    """
    Checks whether the source one is a get descriptor but the other one is not.
    
    Parameters
    ----------
    source_implementation_detail : ``ImplementationDetail``
        Priority source implementation.
    added_implementation_detail : ``ImplementationDetail``
        The added implementation.
    
    Raises
    ------
    TypeError
        - Instance attribute overwritten with other get-property.
    """
    if source_implementation_detail.instance_attribute and source_implementation_detail.implemented:
        if (not added_implementation_detail.instance_attribute):
            raise TypeError(
                f'`{source_implementation_detail.name}` is implemented as instance attribute at '
                f'`{source_implementation_detail.implementation_source}`, but at'
                f'`{added_implementation_detail.implementation_source}` it is not. '
                f'Got `{source_implementation_detail!r}`; `{added_implementation_detail!r}`.'
            )


def _check_method_match(source_implementation_detail, added_implementation_detail):
    """
    Checks whether the source implementation method matches the added implementation one.
    
    Parameters
    ----------
    source_implementation_detail : ``ImplementationDetail``
        Priority source implementation.
    added_implementation_detail : ``ImplementationDetail``
        The added implementation.
    
    Raises
    ------
    TypeError
        - Both method is implemented.
        - Method parameter mismatch.
    """
    if source_implementation_detail.is_method() and added_implementation_detail.is_method():
        if source_implementation_detail.implemented and added_implementation_detail.implemented:
            raise TypeError(
                f'`{source_implementation_detail.name}` is implemented at '
                f'`{source_implementation_detail.implementation_source}` and at '
                f'`{added_implementation_detail.implementation_source}` as well. '
                f'Got `{source_implementation_detail!r}`; `{added_implementation_detail!r}`.'
            )
        
        source_implementation = source_implementation_detail.implementation
        added_implementation = added_implementation_detail.implementation
        
        if (source_implementation is not None) and (added_implementation is not None):
            if not CallableAnalyzer(source_implementation) % CallableAnalyzer(added_implementation):
                raise TypeError(
                    f'`{source_implementation_detail.name}` is implemented differently at '
                    f'`{source_implementation_detail.implementation_source}` than at '
                    f'`{added_implementation_detail.implementation_source}`. '
                    f'Got `{source_implementation_detail!r}`; `{added_implementation_detail!r}`.'
                )


def _check_constant_match(source_implementation_detail, added_implementation_detail):
    """
    Checks for constant variable matching.
    
    Parameters
    ----------
    source_implementation_detail : ``ImplementationDetail``
        Priority source implementation.
    added_implementation_detail : ``ImplementationDetail``
        The added implementation.
    
    Raises
    ------
    TypeError
        - Constant value mismatch.
    """
    if source_implementation_detail.is_constant() and added_implementation_detail.is_constant():
        if source_implementation_detail.implemented and added_implementation_detail.implemented:
            source_implementation = source_implementation_detail.implementation
            added_implementation = added_implementation_detail.implementation
            
            if source_implementation != added_implementation:
                raise TypeError(
                    f'`{source_implementation_detail.name}` is implemented differently at '
                    f'`{source_implementation_detail.implementation_source}` than at '
                    f'`{added_implementation_detail.implementation_source}`. '
                    f'Got `{source_implementation_detail!r}`; `{added_implementation_detail!r}`.'
                )


def _maybe_transfer_instance_attribute_implementation(source_implementation_detail, added_implementation_detail):
    """
    If both implementations are instance attributes and if one is implemented, marks both implemented.
    
    Parameters
    ----------
    source_implementation_detail : ``ImplementationDetail``
        Priority source implementation.
    added_implementation_detail : ``ImplementationDetail``
        The added implementation.
    """
    if source_implementation_detail.instance_attribute and added_implementation_detail.instance_attribute:
        if source_implementation_detail.implemented ^ added_implementation_detail.implemented:
            if source_implementation_detail.implemented:
                added_implementation_detail.implementation_source = source_implementation_detail.implementation_source
            else:
                source_implementation_detail.implementation_source = added_implementation_detail.implementation_source
            
            source_implementation_detail.implemented = True
            added_implementation_detail.implemented = True


def _maybe_transfer_method_implementation(source_implementation_detail, added_implementation_detail):
    """
    If both implementations are methods, transfer the implementation to one from the other if applicable.
    
    Parameters
    ----------
    source_implementation_detail : ``ImplementationDetail``
        Priority source implementation.
    added_implementation_detail : ``ImplementationDetail``
        The added implementation.
    """
    if source_implementation_detail.is_method() and added_implementation_detail.is_method():
        if source_implementation_detail.implemented ^ added_implementation_detail.implemented:
            if source_implementation_detail.implemented:
                added_implementation_detail.implementation_source = source_implementation_detail.implementation_source
                added_implementation_detail.implementation = source_implementation_detail.implementation
            else:
                source_implementation_detail.implementation_source = added_implementation_detail.implementation_source
                source_implementation_detail.implementation = added_implementation_detail.implementation
            
            source_implementation_detail.implemented = True
            added_implementation_detail.implemented = True


def _maybe_transfer_implementation(source_implementation_detail, added_implementation_detail):
    """
    Tries to transfer implementation if only one is implemented.
    
    Parameters
    ----------
    source_implementation_detail : ``ImplementationDetail``
        Priority source implementation.
    added_implementation_detail : ``ImplementationDetail``
        The added implementation.
    """
    if (source_implementation_detail.implemented and source_implementation_detail.implementation is NotImplemented):
        source_implementation_detail.implemented = added_implementation_detail.implemented
        source_implementation_detail.implementation = added_implementation_detail.implementation
        source_implementation_detail.instance_attribute = added_implementation_detail.instance_attribute
    
    elif (added_implementation_detail.implemented and added_implementation_detail.implementation is NotImplemented):
        added_implementation_detail.implemented = source_implementation_detail.implemented
        added_implementation_detail.implementation = source_implementation_detail.implementation
        added_implementation_detail.instance_attribute = source_implementation_detail.instance_attribute



def separate_base_type_from_components(base_types):
    """
    Separates base types from compound components and returns the result.
    
    Parameters
    ----------
    base_types : `tuple` of `type`
        base types to create a new type with.
    
    Returns
    -------
    direct_inherit_type : `None`, `type`
        The directly inherited type.
    components : `None`, `list` of ``CompoundMetaType``
        Compound components.
    
    Raises
    ------
    TypeError
        - If a base-type is a not a type.
        - Only the first type can bo a non-component type.
    """
    components = None
    direct_inherit_type = None
    
    for base_type in base_types:
        if (base_type is object) or (base_type is Compound):
            continue
        
        if not isinstance(base_type, type):
            raise TypeError(
                f'Base types can only be `type` instances, got {base_type!r}; base_types={base_types}.'
            )
        
        if isinstance(base_type, CompoundMetaType):
            if (components is None):
                components = []
            
            components.append(base_type)
        
        else:
            if (direct_inherit_type is None):
                direct_inherit_type = base_type
            
            else:
                raise TypeError(
                    f'Only one type can be non-component base type, got {base_type!r}; base_types={base_types}.'
                )
    
    return direct_inherit_type, components


def build_type_structure(type_name, base_types, type_attributes):
    """
    Builds type structure from the given type constructor parameters.
    
    Parameters
    ----------
    type_name : `str`
        The created type's future name.
    base_types : `tuple` of `type`
        The types to inherit from.
    type_attributes : `dict` of (`str`, `Any`) items
        The attributes defined inside of the type body.
    
    Returns
    -------
    new_base_types : `tuple` of `type`
        The new base types.
    new_type_attributes : `dict` of (`str`, `Any`) items
        The new type attributes.
    
    Raises
    ------
    TypeError
        - If a base-type is a not a type.
        - Only the first type can bo a non-component type.
        - Get-descriptor mismatch.
        - Instance attribute overwritten with other get-property.
        - Both method is implemented.
        - Method parameter mismatch.
        - If a type  attribute is never implemented.
        - Constant value mismatch.
    """
    direct_inherit_type, components = separate_base_type_from_components(base_types)
    
    base_layer, new_type_attributes = CompoundLayer.from_name_and_attributes(type_name, type_attributes)
    logic = CompoundLogicUnit(base_layer)
    
    if (direct_inherit_type is not None):
        logic.add_base_layer(CompoundLayer.from_type(direct_inherit_type, implemented=True))
    
    if (components is not None):
        for component in components:
            logic.add_component_layers(CompoundLayer.from_component(component))
    
    
    type_attribute_implementations = {}
    
    for attribute_name in logic.exhaust_attribute_names():
        implementation_detail = logic.try_get_base_implementation_for(attribute_name)
        
        for layer in logic.component_layers:
            layer_implementation = layer.get_implementation_detail_of(attribute_name, exhaust=True)
            if (layer_implementation is None):
                continue
            
            if implementation_detail is None:
                implementation_detail = layer_implementation
                continue
            
            implementation_detail.apply_layer_implementation(layer_implementation, attribute_name)
            continue
        
        
        if (implementation_detail is None):
            raise TypeError(
                f'`{attribute_name}` is never implemented. Unknown source.'
            )
        
        if not implementation_detail.is_implemented():
            raise TypeError(
                f'`{attribute_name}` is never implemented. Source: {implementation_detail.implementation_source!r}.'
            )
        
        type_attribute_implementations[attribute_name] = implementation_detail
    
    
    implemented_instance_attributes = []
    new_instance_attributes = []
    
    for attribute_name, implementation_detail in type_attribute_implementations.items():
        if not implementation_detail.instance_attribute:
            continue
        
        if implementation_detail.implemented:
            implemented_instance_attributes.append(attribute_name)
        
        else:
            new_instance_attributes.append(attribute_name)
    
    for attribute_name in chain(implemented_instance_attributes, new_instance_attributes):
        del type_attribute_implementations[attribute_name]
    
    
    new_instance_attributes.sort()
    
    new_type_attributes['__slots__'] = tuple(new_instance_attributes)
    
    for attribute_name, implementation_detail in type_attribute_implementations.items():
        new_type_attributes[attribute_name] = implementation_detail.implementation
    
    if direct_inherit_type is None:
        new_base_types = ()
    
    else:
        new_base_types = (direct_inherit_type,)
    
    return new_base_types, new_type_attributes



__all__ = ()

import reprlib

from ..utils import copy_docs

from .headers import CONNECTION, CONTENT_ENCODING, TRANSFER_ENCODING


class RawMessage:
    """
    Base class of ``RawResponseMessage`` and ``RawRequestMessage``.
    
    Attributes
    ----------
    _upgraded : `int`
        Whether the connection is upgraded.
        
        Is only set when the ``.upgraded`` property is accessed as `0`, `1`. Till is set as `2`.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        The headers of the http message.
    """
    __slots__ = ('_upgraded', 'headers', )
    
    def __new__(cls, headers):
        """
        Creates a new raw http message.
        
        Parameters
        ----------
        headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
            The headers of the http message.
        """
        self = object.__new__(cls)
        self._upgraded = 2
        self.headers = headers
        return self
    
    
    @property
    def upgraded(self):
        """
        A get-set descriptor to access whether the message is upgraded. On first access the upgrade state is detected
        from the headers.
        """
        upgraded = self._upgraded
        if upgraded == 2:
            try:
                connection = self.headers[CONNECTION]
            except KeyError:
                upgraded = 0
            else:
                upgraded = (connection.lower() == 'upgrade')
            
            self._upgraded = upgraded
        
        return upgraded
    
    
    @upgraded.setter
    def upgraded(self, value):
        self._upgraded = value
    
    
    @property
    def chunked(self):
        """
        Returns whether the respective http message' body is chunked.
        
        Returns
        -------
        chunked : `bool`
        """
        try:
            transfer_encoding = self.headers[TRANSFER_ENCODING]
        except KeyError:
            return False
        
        return ('chunked' in transfer_encoding.lower())
    
    
    @property
    def encoding(self):
        """
        Returns the body encoding set in the http message's header.
        
        Returns
        -------
        encoding : `None`, `str`
            If no encoding is set, defaults to `None`.
        """
        try:
            encoding = self.headers[CONTENT_ENCODING]
        except KeyError:
            encoding = None
        else:
            encoding = encoding.lower()
        
        return encoding
    
    
    def _cursed_repr_builder(self):
        """
        Helper for ``__repr__`` function.
        
        This method is an iterable generator.
        
        Yields
        ------
        repr_parts : `list` of `str`
        
        Usage
        -----
        ```py
        for repr_parts in self._cursed_repr_builder():
            repr_parts.append(', your fields')
        
        return ''.join(repr_parts)
        ```
        """
        repr_parts = ['<', self.__class__.__name__]
        
        yield repr_parts
        
        repr_parts.append(' headers=')
        repr_parts.append(repr(self.headers))
        
        repr_parts.append(', upgraded=')
        repr_parts.append(repr(self.upgraded))
        
        repr_parts.append(', chunked=')
        repr_parts.append(repr(self.chunked))
        
        encoding = self.encoding
        if (encoding is not None):
            repr_parts.append(', encoding=')
            repr_parts.append(repr(encoding))
        
        repr_parts.append('>')
    
    
    def __repr__(self):
        """Returns the http messages representation."""
        for repr_parts in self._cursed_repr_builder():
            pass
        
        return ''.join(repr_parts)


class RawResponseMessage(RawMessage):
    """
    Represents a raw http response message.
    
    Attributes
    ----------
    _upgraded : `int`
        Whether the connection is upgraded.
        
        Is only set when the ``.upgraded`` property is accessed as `0`, `1`. Till is set as `2`.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        The headers of the http message.
    version : ``HttpVersion``
        The http version of the response.
    status : `int`
        The response's status.
    reason : `bytes`
        Reason included with the response. Might be empty.
    """
    __slots__ = ('version', 'status', 'reason',)
    
    def __new__(cls, version, status, reason, headers):
        """
        Creates a new ``RawResponseMessage`` with the given parameters.
        
        Parameters
        ----------
        version : ``HttpVersion``
            The http version of the response.
        status : `int`
            The response's status.
        reason : `bytes`
            Reason included with the response. Might be empty.
        headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
            The headers of the http message.
        """
        self = RawMessage.__new__(cls, headers)
        
        self.version = version
        self.status = status
        self.reason = reason
        
        return self


    @copy_docs(RawMessage._cursed_repr_builder)
    def _cursed_repr_builder(self):
        for repr_parts in RawMessage._cursed_repr_builder(self):
            
            repr_parts.append(', version=')
            repr_parts.append(repr(self.version))
            
            repr_parts.append(', status=')
            repr_parts.append(repr(self.status))
            
            reason = self.reason
            if reason:
                repr_parts.append(', reason=')
                repr_parts.append(reprlib.repr(self.reason))
            
            yield repr_parts


class RawRequestMessage(RawMessage):
    """
    Represents a raw http request message.
    
    Attributes
    ----------
    _upgraded : `int`
        Whether the connection is upgraded.
        
        Is only set when the ``.upgraded`` property is accessed as `0`, `1`. Till is set as `2`.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        The headers of the http message.
    version : ``HttpVersion``
        The http version of the response.
    method : `int`
        The request's method.
    path : `str`
        The requested path.
    """
    __slots__ = ('version', 'method', 'path',)
    
    def __new__(cls, version, method, path, headers):
        """
        Creates a new ``RawRequestMessage`` with the given parameters.
        
        Parameters
        ----------
        version : ``HttpVersion``
            The http version of the response.
        method : `int`
            The request's method.
        path : `str`
            The requested path.
        headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
            The headers of the http message.
        """
        self = RawMessage.__new__(cls, headers)
        
        self.version = version
        self.method = method
        self.path = path
        
        return self
    
    
    @copy_docs(RawMessage._cursed_repr_builder)
    def _cursed_repr_builder(self):
        for repr_parts in RawMessage._cursed_repr_builder(self):
            
            repr_parts.append(', version=')
            repr_parts.append(repr(self.version))
            
            repr_parts.append(', method=')
            repr_parts.append(repr(self.method))
            
            reason = self.reason
            if reason:
                repr_parts.append(', reason=')
                repr_parts.append(repr(self.path))
            
            yield repr_parts



__all__ = ('MultiValueDictionary', )

from .dict_iterator_bases import DictionaryItemIteratorBase, DictionaryValueIteratorBase
from .docs import copy_docs, has_docs
from .removed_descriptor import RemovedDescriptor
from .utils import is_iterable

@has_docs
class _MultiValueDictionaryItemIterator(DictionaryItemIteratorBase):
    __slots__ = ()
    
    @copy_docs(DictionaryItemIteratorBase.__len__)
    def __len__(self):
        length = 0
        
        for values in dict.values(self._parent):
            length += len(values)
        
        return length
    
    
    @copy_docs(DictionaryItemIteratorBase.__iter__)
    def __iter__(self):
        for key, values in dict.items(self._parent):
            for value in values:
                yield key, value
    
    
    @copy_docs(DictionaryItemIteratorBase.__contains__)
    def __contains__(self, item):
        if not isinstance(item, tuple):
            return False
        
        if len(item) != 2:
            return False
        
        key, value = item
        
        try:
            values = dict.__getitem__(self._parent, key)
        except KeyError:
            return False
        
        return value in values


@has_docs
class _MultiValueDictionaryValueIterator(DictionaryValueIteratorBase):
    __slots__ = ()
    
    @copy_docs(DictionaryValueIteratorBase.__len__)
    def __len__(self):
        length = 0
        
        for values in dict.values(self._parent):
            length += len(values)
        
        return length
    
    
    @copy_docs(DictionaryValueIteratorBase.__iter__)
    def __iter__(self):
        for values in dict.values(self._parent):
            yield from values
    
    
    @copy_docs(DictionaryValueIteratorBase.__contains__)
    def __contains__(self, value):
        for values in dict.values(self._parent):
            if value in values:
                return True
        
        return False


@has_docs
class MultiValueDictionary(dict):
    """
    Dictionary subclass, which can hold multiple values bound to a single key.
    """
    __slots__ = ()
    
    @has_docs
    def __init__(self, iterable=None):
        """
        Creates a new ``MultiValueDictionary``.
        
        Parameters
        ----------
        iterable : `None`, `iterable` = `None`, Optional
            Iterable to update the created dictionary initially.
            
            Can be given as one of the following:
                - ``MultiValueDictionary``.
                - `dict`.
                - `iterable` of `key` - `value` pairs.
        """
        if (iterable is None) or (not iterable):
            return
        
        getitem = dict.__getitem__
        setitem = dict.__setitem__
        
        if isinstance(iterable, MultiValueDictionary):
            for key, values in dict.items(iterable):
                setitem(self, key, values.copy())
        
        elif isinstance(iterable, dict):
            for key, value in iterable.items():
                setitem(self, key, [value])
        
        else:
            for key, value in iterable:
                try:
                    values = getitem(self, key)
                except KeyError:
                    setitem(self, key, [value])
                else:
                    if value not in values:
                        values.append(value)
    
    
    @has_docs
    def __getitem__(self, key):
        """
        Returns the dictionary's `value` for the given `key`. If the `key` has more values, then returns the 0th of
        them.
        """
        return dict.__getitem__(self, key)[0]
    
    
    @has_docs
    def __setitem__(self, key, value):
        """Adds the given `key` - `value` pair to the dictionary."""
        try:
            line = dict.__getitem__(self, key)
        except KeyError:
            dict.__setitem__(self, key, [value])
        else:
            if value not in line:
                line.append(value)
    
    
    @has_docs
    def __delitem__(self, key):
        """
        Removes the `value` for the given `key` from the dictionary. If the `key` has more values, then removes only
        the 0th of them.
        """
        my_list = dict.__getitem__(self, key)
        if len(my_list) == 1:
            dict.__delitem__(self, key)
        else:
            del my_list[0]
    
    
    @has_docs
    def __eq__(self, other):
        """Returns whether the dictionary equals to other."""
        if isinstance(other, type(self)):
            return dict.__eq__(self, other)
        
        if is_iterable(other):
            try:
                other = type(self)(other)
            except TypeError:
                return NotImplemented
            
            return dict.__eq__(self, other)
        
        return NotImplemented
    
    
    @has_docs
    def __ne__(self, other):
        """Returns whether the dictionary equals to other."""
        if isinstance(other, type(self)):
            return dict.__ne__(self, other)
        
        if is_iterable(other):
            try:
                other = type(self)(other)
            except TypeError:
                return NotImplemented
            
            return dict.__ne__(self, other)
        
        return NotImplemented
    
    
    @has_docs
    def __reduce__(self):
        """Reduces the dictionary to a picklable object."""
        return (type(self), list(self.items()))
    
    
    @has_docs
    def __reduce_ex__(self, version):
        """Reduces the dictionary to a picklable object."""
        return type(self).__reduce__(self)
    
    
    @has_docs
    def extend(self, mapping):
        """
        Extends the dictionary with the given `mapping`'s items.
        
        Parameters
        ----------
        mapping : `Any`
            Any mapping type, what has `.items` attribute.
        """
        getitem = dict.__getitem__
        setitem = dict.__setitem__
        
        for key, value in mapping.items():
            try:
                values = getitem(self, key)
            except KeyError:
                setitem(self, key, [value])
            else:
                if value not in values:
                    values.append(value)
    
    
    @has_docs
    def get_all(self, key, default=None):
        """
        Returns all the values matching the given `key`.
        
        Parameters
        ----------
        key : `Any`
            The `key` to match.
        default : `Any` = `None`, Optional
            Default value to return if `key` is not present in the dictionary.
        
        Returns
        -------
        values : `default or `list` of `Any`
            The values for the given `key` if present.
        """
        try:
            return dict.__getitem__(self, key).copy()
        except KeyError:
            return default
    
    
    @has_docs
    def get_one(self, key, default=None):
        """
        Returns the 0th value matching the given `key`.
        
        Parameters
        ----------
        key : `Any`
            The key to match.
        default : `Any` = `None`, Optional
            Default value to return if `key` is not present in the dictionary.
        
        Returns
        -------
        value : `default`, `Any`
            The value for the given key if present.
        """
        try:
            values = dict.__getitem__(self, key)
        except KeyError:
            return default
        else:
            return values[0]
    
    get = get_one
    
    @has_docs
    def setdefault(self, key, default=None):
        """
        Returns the value for the given `key`.
        
        If the `key` is not present in the dictionary, then set's the given `default` value as it.
        
        Parameters
        ----------
        key : `Any`
            The key to match.
        default : `Any` = `None`, Optional
            Default value to set and return if `key` is not present in the dictionary.
        
        Returns
        -------
        value : `default`, `Any`
            The first value for which `key` matched, or `default` if none.
        """
        try:
            values = dict.__getitem__(self, key)
        except KeyError:
            pass
        else:
            return values[0]
        
        dict.__setitem__(self, key, [default])
        return default
    
    
    @has_docs
    def pop_all(self, key, default=...):
        """
        Removes all the values from the dictionary which the given `key` matched.
        
        Parameters
        ----------
        key : `Any`
            The key to match.
        default : `Any`, Optional
            Default value to return if `key` is not present in the dictionary.
        
        Returns
        -------
        values : `default`, `list` of `Any`
            The matched values. If `key` is not present, but `default` value is given, then returns that.
        
        Raises
        ------
        KeyError
            if `key` is not present in the dictionary and `default` value is not given either.
        """
        try:
            return dict.pop(self, key)
        except KeyError:
            if default is not ...:
                return default
            raise
    
    
    @has_docs
    def pop_one(self, key, default=...):
        """
        Removes the first value from the dictionary, which matches the given `key`.
        
        Parameters
        ----------
        key : `Any`
            The key to match.
        default : `Any`, Optional
            Default value to return if `key` is not present in the dictionary.
        
        Returns
        -------
        value : `default`, `list` of `Any`
            The 0th matched value. If `key` is not present, but `default` value is given, then returns that.
        
        Raises
        ------
        KeyError
            if `key` is not present in the dictionary and `default` value is not given either.
        """
        try:
            values = dict.__getitem__(self, key)
        except KeyError:
            if default is not ...:
                return default
            raise
        else:
            value = values.pop(0)
            if not values:
                dict.__delitem__(self, key)
            
            return value
    
    pop = pop_one
    
    
    @has_docs
    def popitem(self):
        """
        Pops an item of the dictionary.
        
        Returns
        -------
        item : `tuple` (`Any`, `Any`)
        
        Raises
        ------
        KeyError
            If the dictionary is empty.
        """
        key, values = dict.popitem(self)
        value = values.pop(0)
        
        if values:
            dict.__setitem__(self, key, values)
        
        return key, value
        
        raise KeyError('popitem(): dictionary is empty.')
    
    
    @has_docs
    def copy(self):
        """
        Copies the dictionary.
        
        Returns
        -------
        new : ``MultiValueDictionary``
        """
        new = dict.__new__(type(self))
        setitem = dict.__setitem__
        for key, values in dict.items(self):
            setitem(new, key, values.copy())
        
        return new
    
    
    @has_docs
    def items(self):
        """
        Returns an item iterator for the dictionary.
        
        Returns
        -------
        items : ``_MultiValueDictionaryItemIterator``
        """
        return _MultiValueDictionaryItemIterator(self)
    
    
    @has_docs
    def values(self):
        """
        Returns a value iterator for the dictionary.
        
        Returns
        -------
        items : ``_MultiValueDictionaryValueIterator``
        """
        return _MultiValueDictionaryValueIterator(self)
    
    
    @has_docs
    def __repr__(self):
        """Returns the dictionary's representation."""
        result = [
            self.__class__.__name__,
            '({',
        ]
        
        if self:
            for key, value in self.items():
                result.append(repr(key))
                result.append(': ')
                result.append(repr(value))
                result.append(', ')
            
            result[-1] = '})'
        else:
            result.append('})')
        
        return ''.join(result)
    
    __str__ = __repr__
    
    @has_docs
    def kwargs(self):
        """
        Converts the dictionary to `**kwargs`-able dictionary. If a `key` has more values, then always returns the last
        value for it.
        
        Returns
        -------
        result : `dict` of (`Any`, `Any`) items
        """
        result = {}
        for key, values in dict.items(self):
            result[key] = values[-1]
        
        return result
    
    update = RemovedDescriptor()



from .event_loop import *
from .protocols_and_transports import *
from .subprocess import *
from .top_level import *
from .traps import *

from .exceptions import *
from .ios import *
from .time import *

__all__ = (
    *event_loop.__all__,
    *protocols_and_transports.__all__,
    *subprocess.__all__,
    *top_level.__all__,
    *traps.__all__,
    
    *exceptions.__all__,
    *ios.__all__,
    *time.__all__,
)



__all__ = ()


def _validate_buffer(buffer):
    """
    Parameters
    ----------
    buffer : `None`, `list` of `str`
        A precreated buffer which commands the initial lines of the input to write.
    
    Returns
    -------
    buffer : `list` of `str`
    
    Raises
    ------
    TypeError
        - If `buffer`'s type is incorrect.
    """
    if (buffer is not None) and (not isinstance(buffer, list)):
        raise TypeError(
            f'buffer can be `None`, `list`, got {buffer.__class__.__name__}; {buffer!r}.'
    )
    
    if buffer is None:
        buffer = []
    
    return buffer


class EditorBase:
    """
    Base editor implementation.
    
    Attributes
    ----------
    compiled_code : `None`, ``CodeType``
        The compiled inputted code.
    file_name : `str`
        File name of the code produced by the editor.
    prefix_initial : `str`
        First line's prefix.
    prefix_continuous : `str`
        Non-first line's prefix.
    prefix_length : `int`
        As how long should the prefix's length be interpreted.
    """
    __slots__ = ('compiled_code', 'file_name', 'prefix_initial', 'prefix_continuous', 'prefix_length')
    
    def __new__(cls, buffer, file_name, prefix_initial, prefix_continuous, prefix_length, highlighter, history):
        """
        Creates a new editor. Initialises with given buffer.
        
        Parameters
        ----------
        buffer : `None`, `list` of `str`
            A precreated buffer which commands the initial lines of the input to write.
        file_name : `str`
            File name of the code produced by the editor.
        prefix_initial : `str`
            First line's prefix.
        prefix_continuous : `str`
            Non-first line's prefix.
        prefix_length : `int`
            As how long should the prefix's length be interpreted.
        highlighter : `None`, ``HighlightFormatterContext``
            Formatter storing highlighting details.
        history : ``History``
            History used for caching inputs.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        """
        self = object.__new__(cls)
        self.compiled_code = None
        self.file_name = file_name
        self.prefix_initial = prefix_initial
        self.prefix_continuous = prefix_continuous
        self.prefix_length = prefix_length
        return self
    
    
    def __repr__(self):
        """Returns the editor's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    def __call__(self):
        """
        Executes the editor's loop.
        
        Returns
        -------
        compiled_code : `None`, ``CodeType``
            the compiled code.
        
        Raises
        ------
        KeyboardInterrupt
        SystemExit
        SyntaxError
        """
        return self.compiled_code
    
    
    def get_buffer(self):
        """
        Returns the editor's buffer.
        
        Returns
        -------
        buffer : `list` of `str`
        """
        return []
    
    
    def is_empty(self):
        """
        Returns whether the editor is empty.
        
        Returns
        -------
        is_empty : `bool`
        """
        for line in self.get_buffer():
            for character in line:
                if character not in {'\n', ' ', '\t', '\r'}:
                    return False
        
        return True
    
    
    def get_stripped_buffer(self):
        """
        Returns the stripped content of the editor.
        
        Returns
        -------
        buffer : `list` of `str`
        """
        lines = [line.rstrip() for line in self.get_buffer()]
        
        while lines and (not lines[0]):
            del lines[0]
        
        while lines and (not lines[-1]):
            del lines [-1]
        
        return lines
    
    
    @classmethod
    def has_history_support(cls):
        """
        Returns whether the editor supports history lookups.
        
        If it does not, it's input should not be saved.
        
        Returns
        -------
        has_history_support : `bool`
        """
        return False



__all__ = ()

import sys, re, termios, tty
from os import get_blocking, get_terminal_size, set_blocking
from select import poll as Poller, POLLOUT as EVENT_POLL_WRITE
from selectors import DefaultSelector, EVENT_READ
from socket import socketpair as create_socket_pair
from ....utils import DEFAULT_ANSI_HIGHLIGHTER, copy_docs, create_ansi_format_code, iter_highlight_code_lines

from .compilation import maybe_compile
from .editor_base import EditorBase, _validate_buffer
from .prefix_trimming import trim_console_prefix


KEY_KEYBOARD_INTERRUPT = 3
KEY_DESTROY = 4
KEY_TAB = 9
KEY_NEW_LINE_ALL = (10, 13)
KEY_ARROW_ALL_INITIAL = 27
KEY_ARROW_ALL = 91
KEY_ARROW_UP = 65
KEY_ARROW_DOWN = 66
KEY_ARROW_RIGHT = 67
KEY_ARROW_LEFT = 68
KEY_ARROW_BACK_TAB = 90
KEY_DELETE_LEFT = 127
KEY_DELETE_RIGHT = 126

COMMAND_NEXT_LINE = '\n'
COMMAND_PREVIOUS_LINE = '\033[F'
COMMAND_START_LINE = '\u001b[1000D'
COMMAND_CLEAR_LINE = '\u001b[0K'
COMMAND_FORMAT_RESET = create_ansi_format_code()

EMPTY_LINE_PREFIX_CHARACTER = ' '
CONTINUOUS_LINE_POSTFIX = '\\'

DEDENT_WORDS = frozenset((
    'pass',
    'return',
    'continue',
    'break'
))

EMPTY_CHARACTERS = frozenset((' ', '\t', '\n'))

INDEXED_INPUT_RP = re.compile('\s*in\s*\[\s*(\d+)\s*\]\s*', re.I)

CHARACTER_CHAINED_OPERATION = chr(KEY_ARROW_ALL_INITIAL)

STDOUT_WRITE_TIMEOUT = 30.0


def create_command_move_cursor(position):
    """
    Creates a command which moves the cursor on the current line to the given position.
    
    Parameters
    ----------
    position : `str`
        The position to move the cursor to
    
    Returns
    -------
    command : `str`
    """
    if position <= 0:
        return ''
    
    return f'\u001b[{position + 1}C'


def get_starting_space_count(line):
    """
    Returns how much space the line has at it's start.
    
    Parameters
    ----------
    line : `str`
        The line to check out.
    
    Returns
    -------
    space_count : `int`
    """
    space_count = 0
    for character in line:
        if character != ' ':
            break
            
        space_count += 1
        continue
    
    return space_count


def add_linebreaks_between(lines):
    """
    Adds linebreaks between the given lines and returns a new list.
    
    Parameters
    ----------
    lines : `list` of `str`
        Lines to add spaces between.
    
    Returns
    -------
    new_lines : `list` of `str`
    """
    new_lines = []
    
    length = len(lines)
    if length:
        
        index = 0
        
        while True:
            line = lines[index]
            new_lines.append(line)
            
            index += 1
            if index >= length:
                break
            
            new_lines.append('\n')
            continue
    
    return new_lines


def line_ends_with(line, target):
    """
    Returns whether the given line ends with the given character.
    
    Spaces ignored.
    
    Parameters
    ----------
    line : `str`
        The lien to check out.
    target : `str`
        The character to look for.
    
    Returns
    -------
    ends_with : `bool`
    """
    for character in reversed(line):
        if character == target:
            return True
        
        if character == ' ':
            continue
        
        break
    
    return False


def line_starts_with_word_any(line, words):
    """
    Returns whether the given line ends with the any of the given words.
    
    Spaces ignored.
    
    Parameters
    ----------
    line : `str`
        The lien to check out.
    words : `frozenset` of `str`
        The words to look for.
    
    Returns
    -------
    ends_with : `bool`
    """
    line = line.lstrip()
    
    for word in words:
        if line.startswith(word):
            if len(line) == len(word) or line[len(word)] == ' ':
                return True
    
    return False


def iter_highlighted_buffer_parts(buffer, highlighter):
    """
    Iterates over the buffer parts and highlights them.
    
    This method is an iterable generator.
    
    Parameters
    ----------
    buffer : `list` of `str`
        Line buffer.
    highlighter : `None`, ``HighlightFormatterContext``
        The highlighter to use.
    
    Yields
    ------
    part : `str`
    """
    buffer = add_linebreaks_between(buffer)
    
    if highlighter is None:
        yield from buffer
    
    else:
        yield from iter_highlight_code_lines(buffer, highlighter)


def is_command(part):
    """
    Returns whether the given part is an ansi command.
    
    Parameters
    ----------
    part : `str`
        the part to check.
    
    Returns
    -------
    is_ansi_command : `bool`
    """
    return part.startswith('\u001b[')


def get_new_content_width(display_state, editor):
    """
    Gets the new column with and whether it changes.
    
    Parameters
    ----------
    display_state : ``DisplayState``
        The current display state.
    editor : ``EditorAdvanced``
        The respective editor.
    
    Returns
    -------
    new_column_with : `int`
        Returns `-1` if there was no change.
    """
    new_column_with = get_terminal_size().columns - editor.prefix_length - len(CONTINUOUS_LINE_POSTFIX)
    if display_state.content_width == new_column_with:
        new_column_with = - 1
    
    return new_column_with


def get_line_line_count(line_length, content_width, new_content_width, prefix_length):
    """
    Returns how much lines the given line is displayed as.
    
    Parameters
    ----------
    line_length : `int`
        The line's length.
    content_width : `int`
        The content field's width.
    new_content_width : `int`
        The new column width to use.
    prefix_length : `int`
        Prefix length used for calculations when column width is changed.
    
    Returns
    -------
    line_line_count : `int`
    """
    if (new_content_width == -1) or (new_content_width > content_width):
        return _get_line_line_count_unchanged(line_length, content_width)
    
    return _get_line_line_count_changed(line_length, content_width, new_content_width, prefix_length)


def _get_line_line_count_unchanged(line_length, content_width):
    """
    Returns how much lines the given line is displayed as.
    
    This function handles the case when content width is not changed or is increased.
    
    Parameters
    ----------
    line_length : `int`
        The line's length.
    content_width : `int`
        The content field's width.
    
    Returns
    -------
    line_line_count : `int`
    """
    if line_length == 0:
        line_line_count = 1
    else:
        line_line_count = 1 + (line_length - 1) // content_width
    
    return line_line_count


def _get_line_line_count_changed(line_length, content_width, new_content_width, prefix_length):
    """
    Returns how much lines the given line is displayed as.
    
    This function handles the case when the content width is reduced and further calculations are required.
    
    Parameters
    ----------
    line_length : `int`
        The line's length.
    content_width : `int`
        The content field's width.
    new_content_width : `int`
        The new column width to use.
    prefix_length : `int`
        Prefix length used for calculations when column width is changed.
    
    Returns
    -------
    line_line_count : `int`
    """
    line_line_count = 0
    
    if line_length == 0:
        line_line_count += 1
    
    else:
        additional_length = prefix_length + len(CONTINUOUS_LINE_POSTFIX)
        old_full_line_length = content_width + additional_length
        new_full_line_length = new_content_width + additional_length
        
        full_line_count = line_length // content_width
        if full_line_count > 0:
            last_line_length = line_length - (full_line_count * content_width)
            
            line_line_count = full_line_count * _get_line_line_count_unchanged(
                old_full_line_length, new_full_line_length
            )
            
        else:
            last_line_length = line_length
        
        if last_line_length > 0:
            line_line_count += 1 + ((last_line_length + prefix_length - 1) // new_full_line_length)
        
    return line_line_count


def _check_is_line_empty(line):
    """
    Returns whether the given line is empty.
    
    Parameters
    ----------
    line : `str`
        The line to check out.
    
    Returns
    -------
    is_empty : `bool`
    """
    for character in line:
        if character not in EMPTY_CHARACTERS:
            return False
    
    return True


def _check_is_line_empty_at(line, index):
    """
    Returns whether the line is empty at the given position.
    
    Parameters
    ----------
    line : `str`
        The line to check out.
    index : `int`
        The exact position.
    
    Returns
    -------
    is_line_empty_at : `bool`
    """
    line_length = len(line)
    if (index < line_length) and (index >= 0):
        if line[index] not in EMPTY_CHARACTERS:
            return False
    
    return True


def _check_is_buffer_line_empty_at(buffer, line_index, index):
    """
    Returns whether the buffer's specified line is empty at the given index.
    
    Parameters
    ----------
    buffer : `list` of `str`
        The buffer to check out.
    line_index : `int`
        The line's index to check out.
    index : `int`
        The exact position.
    
    Returns
    -------
    is_buffer_line_empty_at : `bool`
    """
    buffer_length = len(buffer)
    if (line_index < buffer_length) and (line_index >= 0):
        return _check_is_line_empty_at(buffer[line_index], index)
    
    return True


def _check_is_line_empty_after(line, index):
    """
    Returns whether the line is empty after the given index. Excluding the character at `index` position.
    
    Parameters
    ----------
    line : `str`
        The line to check out.
    index : `int`
        The starting position.
    
    Returns
    -------
    is_line_empty_after : `bool`
    """
    index += 1
    
    line_length = len(line)
    while index < line_length:
        if line[index] not in EMPTY_CHARACTERS:
            return False
        
        index += 1
    
    return True


def _check_is_buffer_line_empty_after(buffer, line_index, index):
    """
    Returns whether the buffer's specified line is empty after the given index.
    Excluding the character at `index` position.
    
    Parameters
    ----------
    buffer : `list` of `str`
        The buffer to check out.
    line_index : `int`
        The line's index to check out.
    index : `int`
        The starting position.
    
    Returns
    -------
    is_buffer_line_empty_after : `bool`
    """
    buffer_length = len(buffer)
    if (line_index < buffer_length) and (line_index >= 0):
        return _check_is_line_empty_after(buffer[line_index], index)
    
    return True


def _check_is_line_empty_before(line, index):
    """
    Returns whether the line is empty before the given index. Excluding the character at `index` position.
    
    Parameters
    ----------
    line : `str`
        The line to check out.
    index : `int`
        The starting position.
    
    Returns
    -------
    is_line_empty_before : `bool`
    """
    index -= 1
    
    line_length = len(line)
    if index >= line_length:
        index = line_length - 1
    
    while index >= 0:
        if line[index] not in EMPTY_CHARACTERS:
            return False
        
        index -= 1
        continue

    return True


def _check_is_buffer_line_empty_before(buffer, line_index, index):
    """
    Returns whether the buffer's specified line is empty before the given index.
    Excluding the character at `index` position.
    
    Parameters
    ----------
    buffer : `list` of `str`
        The buffer to check out.
    line_index : `int`
        The line's index to check out.
    index : `int`
        The starting position.
    
    Returns
    -------
    is_buffer_line_empty_before : `bool`
    """
    buffer_length = len(buffer)
    if (line_index < buffer_length) and (line_index >= 0):
        return _check_is_line_empty_before(buffer[line_index], index)
    
    return True


def _check_is_buffer_empty_at(buffer, index):
    """
    Returns whether the exact line of the buffer is empty.
    
    Parameters
    ----------
    buffer : `list` of `str`
        The buffer to check out.
    index : `int`
        The line's index to check out.
    
    Returns
    -------
    is buffer_empty_at : `bool`
    """
    buffer_length = len(buffer)
    if (index < buffer_length) and (index >= 0):
        return _check_is_line_empty(buffer[index])
    
    return True


def _check_is_buffer_empty_before(buffer, index):
    """
    Returns whether the buffer is empty before the given index. Excluding the `index`'s exact position.
    
    Parameters
    ----------
    buffer : `list` of `str`
        The buffer to check out.
    index : `int`
        The starting position.
    
    Returns
    -------
    is_buffer_empty_before : `bool`
    """
    index -= 1
    
    buffer_length = len(buffer)
    if index >= buffer_length:
        index = buffer_length - 1
    
    
    while index >= 0:
        if not _check_is_line_empty(buffer[index]):
            return False
        
        index -= 1
        continue
    
    return True


def _check_is_buffer_empty_after(buffer, index):
    """
    Returns whether the buffer is empty after the given index. Excluding the `index`'s exact position.
    
    Parameters
    ----------
    buffer : `list` of `str`
        The buffer to check out.
    index : `int`
        The starting position.
    
    Returns
    -------
    is_buffer_empty_after : `bool`
    """
    buffer_length = len(buffer)
    
    index += 1
    
    while index < buffer_length:
        if not _check_is_line_empty(buffer[index]):
            return False
        
        index += 1
        continue
    
    return True


def _check_should_try_compile_one_line(display_state):
    """
    Checks whether the current input should be tried to be compiled.
    
    This function is called when the buffer's length is `exactly 1`.
    
    Parameters
    ----------
    display_state : ``DisplayState``
        The respective input.
    
    Returns
    -------
    should_try_compile : `bool`
    """
    line = display_state.buffer[display_state.cursor_line_index]
    cursor_index = display_state.cursor_index
    
    if not _check_is_line_empty_after(line, cursor_index):
        return False
    
    if not _check_is_line_empty_at(line, cursor_index):
        return False
    
    if _check_is_line_empty_before(line, cursor_index):
        return False
    
    return True


def _check_should_try_compile_multi_line(display_state):
    """
    Checks whether the current input should be tried to be compiled.
    
    This function is called when the buffer's length is `over 1`.
    
    Parameters
    ----------
    display_state : ``DisplayState``
        The respective input.
    
    Returns
    -------
    should_try_compile : `bool`
    """
    buffer = display_state.buffer
    
    # If we are checking multi-line
    
    if len(buffer) < 3:
        return False
    
    cursor_index = display_state.cursor_index
    cursor_line_index = display_state.cursor_line_index
    
    # If there is anything after our current position return false.
    
    if (
        not _check_is_buffer_line_empty_at(buffer, cursor_line_index, cursor_index) or
        not _check_is_buffer_line_empty_after(buffer, cursor_line_index, cursor_index) or
        not _check_is_buffer_empty_after(buffer, cursor_line_index)
    ):
        return False
    
    # If there is no code before the current position return false.
    
    if (
        _check_is_buffer_line_empty_before(buffer, cursor_line_index, cursor_index) and
        _check_is_buffer_empty_before(buffer, cursor_line_index)
    ):
        return False
    
    # We should only compile when we have 1 free line above us and our current line is empty too.
    if (
        _check_is_buffer_empty_after(buffer, cursor_line_index) and
        _check_is_buffer_empty_at(buffer, cursor_line_index) and
        _check_is_buffer_empty_at(buffer, cursor_line_index - 1)
    ):
        return True
    
    return False


def _check_should_try_compile(display_state):
    """
    Tries whether based on the current input, we should try to compile the input.
    
    Parameters
    ----------
    display_state : ``DisplayState``
        The respective input.
    
    Returns
    -------
    should_try_compile : `bool`
    """
    buffer_length = len(display_state.buffer)
    
    if buffer_length == 0:
        # Should not happen
        return False
    
    if buffer_length == 1:
        return _check_should_try_compile_one_line(display_state)
    
    return _check_should_try_compile_multi_line(display_state)


class DisplayState:
    """
    A current display state of ``EditorAdvanced``.
    
    Attributes
    ----------
    buffer : `list` of `str`
        Line buffer.
    content_width : `int`
        The screen's with.
    cursor_index : `int`
        The cursors index on the current line.
    cursor_line_index : `int`
        The cursor's line's index in the buffer.
    """
    __slots__ = ('buffer', 'content_width', 'cursor_index', 'cursor_line_index')
    
    def __new__(cls, buffer):
        """
        Creates a new display state.
        
        Parameters
        ----------
        buffer : `None`, `list` of `str`
            Line buffer.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        """
        buffer = _validate_buffer(buffer)
        if not buffer:
            buffer.append('')
        
        self = object.__new__(cls)
        self.buffer = buffer
        self.cursor_index = len(buffer[-1])
        self.cursor_line_index = len(buffer) - 1
        self.content_width = -1
        return self
    
    
    def __repr__(self):
        """
        Returns the representation of the display state.
        """
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' cursor_index=')
        repr_parts.append(repr(self.cursor_index))
        
        repr_parts.append(', cursor_line_index=')
        repr_parts.append(repr(self.cursor_line_index))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def copy(self):
        """
        Copies the display state.
        
        Returns
        -------
        new : ``DisplayState``
        """
        new = object.__new__(type(self))
        new.buffer = self.buffer.copy()
        new.cursor_index = self.cursor_index
        new.cursor_line_index = self.cursor_line_index
        new.content_width = -1
        return new
    
    
    def get_cursor_till_end_display_line_count(self, new_content_width, prefix_length):
        """
        Counts how much lines are from the display position till the end of the display.
        
        Parameters
        ----------
        new_content_width : `int`
            The new column width to use.
        prefix_length : `int`
            Prefix length used for calculations when column width is changed.
        
        Returns
        -------
        count : `int`
        """
        buffer = self.buffer
        cursor_index = self.cursor_index
        cursor_line_index = self.cursor_line_index
        content_width = self.content_width
        
        line = buffer[cursor_line_index]
        line_length = len(line)
        
        line_line_count = get_line_line_count(line_length, content_width, new_content_width, prefix_length)
        line_cursor_index = get_line_line_count(cursor_index, content_width, new_content_width, prefix_length)
        count = line_line_count - line_cursor_index
        
        index = cursor_line_index + 1
        limit = len(buffer)
        if index < limit:
            while True:
                count += get_line_line_count(len(buffer[index]), content_width, new_content_width, prefix_length)
                
                index += 1
                if index == limit:
                    break
                
                continue
        
        return count
    
    
    def get_start_till_end_display_line_count(self, new_content_width, prefix_length):
        """
        Counts how much lines are displayed.
        
        Parameters
        ----------
        new_content_width : `int`
            The new column width to use.
        prefix_length : `int`
            Prefix length used for calculations when column width is changed.
        
        Returns
        -------
        count : `int`
        """
        buffer = self.buffer
        count = 0
        content_width = self.content_width
        
        for line in buffer:
            count += get_line_line_count(len(line), content_width, new_content_width, prefix_length)
        
        return count
    
    
    def jump_to_end(self, editor):
        """
        Jumps to end of the displayed content.
        
        Parameters
        ----------
        editor : ``EditorAdvanced``
            The editor to use.
        """
        if (self.content_width == -1):
            # self was never rendered
            return
        
        self._jump_to_end(editor, get_new_content_width(self, editor))
    
    
    def _jump_to_end(self, editor, new_content_width):
        """
        Jumps to end of the displayed content.
        
        Called by ``.jump_to_end`` after checking pre-requirements.
        
        Parameters
        ----------
        editor : ``EditorAdvanced``
            The editor
        new_content_width : `int`
            The new column width to use.
        """
        output_stream = editor.output_stream
        
        for _ in range(self.get_cursor_till_end_display_line_count(new_content_width, editor.prefix_length)):
            write_to_io(output_stream, COMMAND_NEXT_LINE)
    
    
    def clear(self, editor):
        """
        Clears self from the editor.
        
        Parameters
        ----------
        editor : ``EditorAdvanced``
            The editor to use.
        """
        if (self.content_width == -1):
            # self was never rendered
            return
        
        new_content_width = get_new_content_width(self, editor)
        
        self._jump_to_end(editor, new_content_width)
        self._clear_from_end(editor, new_content_width)
    
    
    def _clear_from_end(self, editor, new_content_width):
        """
        Clears self from the editor. Should be called after ``._jump_to_end``.
        
        Parameters
        ----------
        editor : ``EditorAdvanced``
            The editor to use.
        new_content_width : `int`
            The new column width to use.
        """
        output_stream = editor.output_stream
        
        write_to_io(output_stream, COMMAND_START_LINE)
        write_to_io(output_stream, COMMAND_CLEAR_LINE)
        
        for _ in range(self.get_start_till_end_display_line_count(new_content_width, editor.prefix_length) - 1):
            write_to_io(output_stream, COMMAND_PREVIOUS_LINE)
            write_to_io(output_stream, COMMAND_START_LINE)
            write_to_io(output_stream, COMMAND_CLEAR_LINE)
    
    
    def write_cursor(self, editor):
        """
        Jumps the cursor to it's location.
        
        The output must be written out before.
        
        Parameters
        ----------
        editor : ``EditorAdvanced``
            The editor to use.
        """
        # Jump back to cursor line
        output_stream = editor.output_stream
        
        for _ in range(self.get_cursor_till_end_display_line_count(-1, editor.prefix_length)):
            write_to_io(output_stream, COMMAND_PREVIOUS_LINE)
        
        write_to_io(output_stream, COMMAND_START_LINE)
        
        cursor_index = self.cursor_index
        if cursor_index == 0:
            line_cursor_index = -1
        else:
            line_cursor_index = ((cursor_index - 1) % self.content_width)
        
        line_cursor_index += editor.prefix_length
        
        write_to_io(output_stream, create_command_move_cursor(line_cursor_index))
    
    
    def write(self, editor):
        """
        Writes out the buffer.
        
        Parameters
        ----------
        editor : ``EditorAdvanced``
            The editor to use.
        """
        buffer = self.buffer
        output_stream = editor.output_stream
        
        prefix_initial = editor.prefix_initial
        prefix_continuous = editor.prefix_continuous
        prefix_length = editor.prefix_length
        
        content_width = get_terminal_size().columns - prefix_length - len(CONTINUOUS_LINE_POSTFIX)
        
        if content_width < 1:
            return
        
        self.content_width = content_width
        
        if len(buffer) == 1 and not buffer[0]:
            write_to_io(output_stream, COMMAND_START_LINE)
            write_to_io(output_stream, COMMAND_CLEAR_LINE)
            write_to_io(output_stream, COMMAND_FORMAT_RESET)
            write_to_io(output_stream, prefix_initial)
            return
        
        line_index = 0
        line_count = len(buffer)
        is_new_line = True
        leftover_line_length = content_width
        
        for part in iter_highlighted_buffer_parts(buffer, DEFAULT_ANSI_HIGHLIGHTER):
            
            if part.endswith('\n'):
                is_line_break = True
                part = part[:-2]
            else:
                is_line_break = False
            
            if is_new_line:
                leftover_line_length = content_width
                
                write_to_io(output_stream, COMMAND_START_LINE)
                write_to_io(output_stream, COMMAND_CLEAR_LINE)
                write_to_io(output_stream, COMMAND_FORMAT_RESET)
                
                is_new_line = False
                
                if line_index == 0:
                    prefix = prefix_initial
                else:
                    prefix = prefix_continuous
                
                write_to_io(output_stream, prefix)
            
            if part:
                if is_command(part):
                    write_to_io(output_stream, part)
                
                else:
                    while True:
                        part_length = len(part)
                        leftover_line_length -= part_length
                        if leftover_line_length >= 0:
                            write_to_io(output_stream, part)
                            break
                        
                        write_to_io(output_stream, part[:leftover_line_length])
                        part = part[leftover_line_length:]
                        
                        write_to_io(output_stream, CONTINUOUS_LINE_POSTFIX)
                        write_to_io(output_stream, COMMAND_NEXT_LINE)
                        write_to_io(output_stream, COMMAND_START_LINE)
                        write_to_io(output_stream, EMPTY_LINE_PREFIX_CHARACTER * prefix_length)
                        leftover_line_length = content_width
                        continue
            
            if is_line_break:
                # At last line skip line break
                line_index += 1
                if line_index == line_count:
                    break
                
                is_new_line = True
                write_to_io(output_stream, COMMAND_NEXT_LINE)
                write_to_io(output_stream, COMMAND_START_LINE)
                write_to_io(output_stream, prefix_continuous)
                continue
            
            continue


def _get_input_content_length(content):
    """
    Gets the input content's length.
    
    Parameter
    ---------
    content : `str`
        Input content.
    
    Returns
    -------
    length : `int`
    """
    length = len(content)
    if length:
        length -= (content.count(CHARACTER_CHAINED_OPERATION) << 1)
        
        if length < 1:
            length = 1
    
    return length


class InputIterator:
    """
    Helper class for processing a bigger chunk of input data at once.
    
    Attributes
    ----------
    iterator : `None, `str_iterator`
        Content iterator.
    length : `int`
        The estimated length of the input.
    """
    __slots__ = ('iterator', 'length')
    
    def __new__(cls, content):
        """
        Creates a new input iterator.
        
        Parameters
        ----------
        content : `str`
            The content to iterate over.
        """
        length = _get_input_content_length(content)
        if length:
            trimmed_content = trim_console_prefix(content)
            if (trimmed_content is not None):
                content = trimmed_content
                length = _get_input_content_length(content)
        
        if length:
            iterator = iter(content)
        else:
            iterator = None
        
        self = object.__new__(cls)
        self.iterator = iterator
        self.length = length
        return self
    
    
    def __iter__(self):
        """Returns self."""
        return self
    
    
    def __next__(self):
        """Gets the next element of the input iterator."""
        iterator = self.iterator
        if iterator is not None:
            next_ = next(iterator, None)
            if next_ is not None:
                return next_
            
            self.iterator = None
        
        raise StopIteration()
    
    
    def get_next(self):
        """Manually reads the next element."""
        iterator = self.iterator
        if iterator is not None:
            next_ = next(iterator, None)
            if next_ is not None:
                return next_
            
            self.iterator = None
        
        return None


def write_to_io(io, content):
    """
    Writes to the given `io`
    
    Parameters
    ----------
    io : `file-like`
        The io to write to.
    content : `str`
        The content to write.
    
    Raises
    ------
    RuntimeError
        - If the `io` did not became writable within timeout.
    """
    file_descriptor = io.fileno()
    if not get_blocking(file_descriptor):
        poller = Poller()
        poller.register(file_descriptor, EVENT_POLL_WRITE)
        events = poller.poll(STDOUT_WRITE_TIMEOUT)
        if not events:
            raise RuntimeError(
                f'The io did not became writable within timeout ({STDOUT_WRITE_TIMEOUT}).'
            )
        
    io.write(content)


class EditorAdvanced(EditorBase):
    """
    A simple repl command editor.
    
    Attributes
    ----------
    alive : `str`
        Whether the input loop is alive.
    async_output_written : `bool`
        Whether async output was already written.
    compiled_code : `None`, ``CodeType``
        The compiled inputted code.
    display_state : ``DisplayState``
        The currently displayed state.
    file_name : `str`
        File name of the code produced by the editor.
    highlighter : `None`, ``HighlightFormatterContext``
        Formatter storing highlighting details.
    history : ``History``
        History used for caching inputs.
    input_stream : `file-like`
        Input stream to read.
    input_stream_blocking_original : `bool`
        The original value of input stream's blockedness.
    input_stream_settings_original : `list` (`int`, `int`, `int`, `int`, `int`, `int`, `list` of `bytes`)
        The original settings og the input stream.
    modified : `bool`
        Whether the input is modified.
    output_proxy_read_socket : `Socket`
        Read pair socket of ``.output_proxy_write_socket``.
    output_proxy_write_socket : `Socket`
        Socket proxying stdout and stderr.
    output_stream : `file-like`
        Output stream to write.
    prefix_continuous : `str`
        Non-first line's prefix.
    prefix_initial : `str`
        First line's prefix.
    prefix_length : `int`
        As how long should the prefix's length be interpreted
    selector : ``DefaultSelector``
        Selector used to poll from `stdin`, `stdout` and `stderr`.
    """
    __slots__ = (
        'alive', 'async_output_written', 'display_state', 'file_name', 'highlighter', 'history', 'input_stream',
        'input_stream_blocking_original', 'input_stream_settings_original', 'modified', 'output_proxy_read_socket',
        'output_proxy_write_socket', 'output_stream', 'selector'
    )
    
    @copy_docs(EditorBase.__new__)
    def __new__(cls, buffer, file_name, prefix_initial, prefix_continuous, prefix_length, highlighter, history):
        
        display_state = DisplayState(buffer)
        
        input_stream = sys.stdin
        input_stream_file_descriptor = input_stream.fileno()
        input_stream_blocking_original = get_blocking(input_stream_file_descriptor)
        set_blocking(input_stream_file_descriptor, False)
        input_stream_settings_original = termios.tcgetattr(input_stream)
        
        output_proxy_read_socket, output_proxy_write_socket = create_socket_pair()
        set_blocking(output_proxy_read_socket.fileno(), False)
        
        
        self = EditorBase.__new__(
            cls, buffer, file_name, prefix_initial, prefix_continuous, prefix_length, highlighter, history
        )
        
        self.alive = False
        self.async_output_written = False
        self.display_state = display_state
        self.history = history
        self.input_stream = input_stream
        self.input_stream_blocking_original = input_stream_blocking_original
        self.input_stream_settings_original = input_stream_settings_original
        self.modified = False
        self.output_proxy_read_socket = output_proxy_read_socket
        self.output_proxy_write_socket = output_proxy_write_socket
        self.output_stream = sys.stdout
        self.selector = DefaultSelector()
        
        return self
    
    
    @copy_docs(EditorBase.__call__)
    def __call__(self):
        new_stdout = open(self.output_proxy_write_socket.fileno(), 'w')
        old_stderr = sys.stderr
        sys.stdout = new_stdout
        sys.stderr = new_stdout
        
        tty.setraw(self.input_stream)
        self.selector.register(self.input_stream.fileno(), EVENT_READ)
        self.selector.register(self.output_proxy_read_socket.fileno(), EVENT_READ)
        self.output_stream.flush()
        
        try:
            self.alive = True
            self.initialise_display()
            
            while True:
                if not self.alive:
                    break
                
                new_display_state = self.poll_and_process_input()
                if (new_display_state is not None):
                    self.update_display(new_display_state)
        
        finally:
            self.alive = False
            self.display_state.jump_to_end(self)
            write_to_io(self.output_stream, '\n')
            write_to_io(self.output_stream, COMMAND_START_LINE)
            write_to_io(self.output_stream, COMMAND_FORMAT_RESET)
            self.output_stream.flush()
            
            termios.tcsetattr(self.input_stream, termios.TCSADRAIN, self.input_stream_settings_original)
            set_blocking(self.input_stream.fileno(), self.input_stream_blocking_original)
            
            sys.stdout = self.output_stream
            sys.stderr = old_stderr
            
            try:
                self.output_proxy_read_socket.close()
            except OSError as err:
                if err.errno != 9:
                    raise
            finally:
                self.output_proxy_read_socket = None
            
            try:
                self.output_proxy_write_socket.close()
            except OSError as err:
                if err.errno != 9:
                    raise
            finally:
                self.output_proxy_write_socket = None
            
            self.selector.close()
        
        return self.compiled_code
    
    
    @copy_docs(EditorBase.get_buffer)
    def get_buffer(self):
        return self.display_state.buffer
    
    
    @classmethod
    @copy_docs(EditorBase.has_history_support)
    def has_history_support(cls):
        return True
    
    
    def execute_enter(self, old_display_state, should_auto_format):
        """
        Executes an enter key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        should_auto_format : `bool`
            Whether auto formatting should be applied.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        
        Raises
        ------
        SyntaxError
        """
        if should_auto_format:
            new_display_state = self.checkout_indexed_input(old_display_state)
            if (new_display_state is not None):
                return new_display_state
            
            if self.check_exit_conditions(old_display_state):
                return None
        
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        line = buffer[cursor_line_index]
        current_line_new = line[:cursor_index]
        next_line = line[cursor_index:]
        
        if should_auto_format:
            new_line_starter_space_count = get_starting_space_count(line)
            # In which order should we check this?
            if line_ends_with(current_line_new, ':'):
                new_line_starter_space_count += 4
            elif line_starts_with_word_any(current_line_new, DEDENT_WORDS):
                new_line_starter_space_count -= 4
            
            if new_line_starter_space_count < 0:
                new_line_starter_space_count = 0
            
            next_line = ' ' * new_line_starter_space_count + next_line
        
        else:
            new_line_starter_space_count = 0
        
        buffer[cursor_line_index] = current_line_new
        cursor_line_index += 1
        buffer.insert(cursor_line_index, next_line)
        
        new_display_state.cursor_index = new_line_starter_space_count
        new_display_state.cursor_line_index = cursor_line_index
        
        self.modified = True
        
        return new_display_state
    
    
    def execute_arrow_left(self, old_display_state):
        """
        Executes an arrow-left key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        if (cursor_index == 0):
            if (cursor_line_index == 0):
                return None
            
            cursor_line_index -= 1
            cursor_index = len(buffer[cursor_line_index])
        
        else:
            cursor_index -= 1
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        return new_display_state
    
    
    def execute_arrow_right(self, old_display_state):
        """
        Executes an arrow-right key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        line_length = len(buffer[cursor_line_index])
        if cursor_index == line_length:
            if cursor_line_index == len(buffer) - 1:
                return None
            
            cursor_line_index += 1
            cursor_index = 0
        else:
            cursor_index += 1
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        return new_display_state
    
    
    def execute_arrow_up(self, old_display_state, should_auto_format):
        """
        Executes an arrow-up key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        should_auto_format : `bool`
            Whether the content should be auto formatted.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        if should_auto_format:
            new_display_state = self.try_rewind_in_history()
            if (new_display_state is not None):
                return new_display_state
        
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        if cursor_line_index == 0:
            if (cursor_index == 0):
                return None
            
            cursor_index = 0
        
        else:
            cursor_line_index -= 1
            
            line_length = len(buffer[cursor_line_index])
            if line_length < cursor_index:
                cursor_index = line_length
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        return new_display_state
    
    
    def execute_arrow_down(self, old_display_state, should_auto_format):
        """
        Executes an arrow-down key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        should_auto_format : `bool`
            Whether the content should be auto formatted.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        if should_auto_format:
            new_display_state = self.try_forward_in_history()
            if (new_display_state is not None):
                return new_display_state
        
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        if cursor_line_index == len(buffer) - 1:
            line_length = len(buffer[cursor_line_index])
            if cursor_index == line_length:
                return None
            
            cursor_index = line_length
        
        else:
            cursor_line_index += 1
            
            line_length = len(buffer[cursor_line_index])
            if line_length < cursor_index:
                cursor_index = line_length
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        return new_display_state
    
    
    def execute_back_tab(self, old_display_state):
        """
        Executes a back-tab key press (shift + tab).
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        line = buffer[cursor_line_index]
        
        back_space_count = get_starting_space_count(line)
        if back_space_count == 0:
            return None
        
        remove_space_count = 4 - (-back_space_count & 3)
        cursor_index -= remove_space_count
        if (cursor_index < 0):
            cursor_index = 0
        line = line[remove_space_count:]
        buffer[cursor_line_index] = line
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        self.modified = True
        
        return new_display_state
    
    
    def execute_tab(self, old_display_state, should_auto_format):
        """
        Executes a tab key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        should_auto_format : `bool`
            Whether auto formatting should be applied.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        line = buffer[cursor_line_index]
        
        if should_auto_format:
            add_spaces_count =  4 - (cursor_index & 3)
            line_length = len(line)
            while add_spaces_count:
                if cursor_index >= line_length:
                    break
                
                if line[cursor_index] != ' ':
                    break
                
                cursor_index += 1
                add_spaces_count -= 1
        else:
            add_spaces_count = 4
        
        if add_spaces_count:
            buffer[cursor_line_index] = f'{line[:cursor_index]}{" " * add_spaces_count}{line[cursor_index:]}'
            cursor_index += add_spaces_count
        
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        self.modified = True
        
        return new_display_state
    
    
    def execute_delete_left(self, old_display_state):
        """
        Executes a delete-left key press (backspace).
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        if cursor_index == 0:
            if cursor_line_index == 0:
                return None
                
            line = buffer.pop(cursor_line_index)
            cursor_line_index -= 1
            previous_line = buffer[cursor_line_index]
            buffer[cursor_line_index] = previous_line + line
            cursor_index = len(previous_line)
        
        else:
            line = buffer[cursor_line_index]
            buffer[cursor_line_index] = f'{line[:cursor_index - 1]}{line[cursor_index:]}'
            cursor_index -= 1
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        self.modified = True
        
        return new_display_state
    
    
    def execute_delete_right(self, old_display_state):
        """
        Executes a delete-right key press (delete).
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        line = buffer[cursor_line_index]
        line_length = len(line)
        if cursor_index == line_length:
            if cursor_line_index == len(buffer) - 1:
                return None
                
            next_line = buffer.pop(cursor_line_index + 1)
            line = line + next_line
            buffer[cursor_line_index] = line
                
        else:
            line = f'{line[:cursor_index]}{line[cursor_index + 1:]}'
            buffer[cursor_line_index] = line
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        self.modified = True
        
        return new_display_state
    
    
    def execute_print(self, old_display_state, character_string):
        """
        Executes a printable character key press.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        character_string : `str`
            The character to print out.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        new_display_state = old_display_state.copy()
        buffer = new_display_state.buffer
        cursor_index = new_display_state.cursor_index
        cursor_line_index = new_display_state.cursor_line_index
        
        line = buffer[cursor_line_index]
        buffer[cursor_line_index] = f'{line[:cursor_index]}{character_string}{line[cursor_index:]}'
        cursor_index += 1
        
        new_display_state.cursor_index = cursor_index
        new_display_state.cursor_line_index = cursor_line_index
        
        self.modified = True
        
        return new_display_state
    
    
    def poll(self):
        """
        Polls a character from any input stream.
        
        Returns
        -------
        content : `str`
            The inputted string into `sys.stdin`.
        """
        content = None
        
        while True:
            # We force flush every second.
            sys.stdout.flush()
            
            event_list = self.selector.select(1)
            if not event_list:
                continue
                
            for key, mask in event_list:
                file_descriptor = key.fileobj
                
                if (file_descriptor == self.output_proxy_read_socket.fileno()) and (mask & EVENT_READ):
                    try:
                        data = self.output_proxy_read_socket.recv(1024)
                    except (BlockingIOError, InterruptedError):
                        pass
                    else:
                        if data:
                            self._display_async_content(data.decode())
                        
                        data = None
                    
                    continue
                
                if (file_descriptor == self.input_stream.fileno()) and (mask & EVENT_READ):
                    try:
                        content = self.input_stream.read()
                    except (BlockingIOError, InterruptedError):
                        pass
                    
                    continue
            
            
            if (content is not None):
                if content:
                    return content
                
                content = None
    
    
    def poll_and_process_input(self):
        """
        Inputs one action and returns whether any actions took place.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        
        Raises
        ------
        KeyboardInterrupt
        SystemExit
        SyntaxError
        """
        content = self.poll()
        input_iterator = InputIterator(content)
        should_auto_format = input_iterator.length <= 1
        
        old_display_state = self.display_state
        new_display_state = old_display_state
        
        for character_string in input_iterator:
            received_display_state = self.process_input(
                new_display_state, input_iterator, character_string, should_auto_format
            )
            if (received_display_state is not None):
                new_display_state = received_display_state
        
        if (new_display_state is old_display_state):
            new_display_state = None
        
        return new_display_state
    
    
    def process_input(self, old_display_state, input_iterator, character_string, should_auto_format):
        """
        Processes one input.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        input_iterator : ``InputIterator``
            The input iterator to pull additional characters if required.
        character_string : `str`
            The input string.
        should_auto_format : `bool`
            Whether the content should be auto formatted.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        character_int = ord(character_string)
        
        if character_int == KEY_KEYBOARD_INTERRUPT:
            raise KeyboardInterrupt()
        
        if character_int == KEY_DESTROY:
            raise SystemExit()
        
        if character_int in KEY_NEW_LINE_ALL:
            return self.execute_enter(old_display_state, should_auto_format)
        
        if character_int == KEY_ARROW_ALL_INITIAL:
            next_1_string = input_iterator.get_next()
            next_2_string = input_iterator.get_next()
            
            if (next_1_string is None) or (next_2_string is None):
                # Escape was pressed.
                return None
            
            next_1_int = ord(next_1_string)
            next_2_int = ord(next_2_string)
            
            if next_1_int != KEY_ARROW_ALL:
                return None
            
            if next_2_int == KEY_ARROW_LEFT:
                return self.execute_arrow_left(old_display_state)
            
            if next_2_int == KEY_ARROW_RIGHT:
                return self.execute_arrow_right(old_display_state)
            
            if next_2_int == KEY_ARROW_UP:
                return self.execute_arrow_up(old_display_state, should_auto_format)
            
            if next_2_int == KEY_ARROW_DOWN:
                return self.execute_arrow_down(old_display_state, should_auto_format)
            
            if next_2_int == KEY_ARROW_BACK_TAB:
                return self.execute_back_tab(old_display_state)
            
            return None
        
        if character_int == KEY_TAB:
            return self.execute_tab(old_display_state, should_auto_format)
        
        if character_int == KEY_DELETE_LEFT:
            return self.execute_delete_left(old_display_state)
        
        if character_int == KEY_DELETE_RIGHT:
            return self.execute_delete_right(old_display_state)
        
        if character_string.isprintable():
            return self.execute_print(old_display_state, character_string)
        
        return None
    
    
    def initialise_display(self):
        """
        Initialises the display.
        
        Called after the editor was started.
        """
        self.display_state.write(self)
        
        self.output_stream.flush()
    
    
    def update_display(self, new_display_state):
        """
        Updates the display.
        
        This method is called after an action took place.
        
        Parameters
        ----------
        new_display_state : ``DisplayState``
            The new display state to render.
        """
        old_display_state = self.display_state
        self.display_state = new_display_state
        
        old_display_state.clear(self)
        new_display_state.write(self)
        new_display_state.write_cursor(self)
        
        self.output_stream.flush()
    
    
    def _display_async_content(self, content):
        """
        Writes the given data before the editor.
        
        Parameters
        ----------
        content : `str`
            The content to write.
        """
        self.display_state.clear(self)
        
        async_output_written = self.async_output_written
        
        if async_output_written:
            write_to_io(self.output_stream, COMMAND_PREVIOUS_LINE)
        
        termios.tcsetattr(self.input_stream, termios.TCSADRAIN, self.input_stream_settings_original)
        write_to_io(self.output_stream, content)
        tty.setraw(self.input_stream)
        
        write_to_io(self.output_stream, COMMAND_NEXT_LINE)
        write_to_io(self.output_stream, COMMAND_START_LINE)
        
        if not async_output_written:
            self.async_output_written = True
        
        self.display_state.write(self)
        self.display_state.write_cursor(self)
        
        self.output_stream.flush()
    
    
    def check_exit_conditions(self, old_display_state):
        """
        Checks whether any exit conditions are met.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Raises
        ------
        SyntaxError
        """
        if _check_should_try_compile(old_display_state):
            compiled_code = maybe_compile(old_display_state.buffer, self.file_name)
            if (compiled_code is not None):
                self.compiled_code = compiled_code
                self.alive = False
                return True
        
        return False
    
    
    def should_move_in_history(self):
        """
        Returns whether history moving can be executed.
        """
        display_state = self.display_state
        if display_state.cursor_line_index != 0:
            return False
        
        display_state = self.display_state
        if display_state.cursor_index != 0:
            return False
        
        return self.can_move_in_history()
        
    
    def can_move_in_history(self):
        """
        Returns whether history moving can be executed.
        
        Returns
        -------
        can_move_in_history : `bool`
        """
        if not self.modified:
            return True
        
        if self.is_empty():
            self.modified = False
            return True
        
        return False
    
    
    def try_rewind_in_history(self):
        """
        Tries to go back in history if applicable.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        if not self.should_move_in_history():
            return None
        
        buffer = self.history.get_previous()
        if buffer is None:
            buffer = ['']
        
        display_state = self.display_state.copy()
        display_state.buffer = buffer.copy()
        return display_state
    
    
    def try_forward_in_history(self):
        """
        Tries to go forward in history if applicable.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        if not self.should_move_in_history():
            return None
        
        buffer = self.history.get_next()
        if buffer is None:
            buffer = ['']
        
        display_state = self.display_state.copy()
        display_state.buffer = buffer.copy()
        return display_state
    
    
    def checkout_indexed_input(self, old_display_state):
        """
        Checks out whether an index of the history. If it is returns a new display state.
        
        Parameters
        ----------
        old_display_state : ``DisplayState``
            The old display state to work from.
        
        Returns
        -------
        new_display_state : `None`, ``DisplayState``
            The new display state.
        """
        index = -1
        
        for line in old_display_state.buffer:
            if _check_is_line_empty(line):
                continue
            
            if index != -1:
                return None
            
            matched = INDEXED_INPUT_RP.fullmatch(line)
            if matched is None:
                return None
            
            index = int(matched.group(1))
            continue
        
        if index == -1:
            return None
        
        buffer = self.history.get_at(index)
        return DisplayState(buffer)



__all__ = ('WebSocketFrame', )

from .exceptions import WebSocketProtocolError


WEBSOCKET_OPERATION_CONTINUOUS = 0
WEBSOCKET_OPERATION_TEXT = 1
WEBSOCKET_OPERATION_BINARY = 2

WEBSOCKET_OPERATION_CLOSE = 8
WEBSOCKET_OPERATION_PING = 9
WEBSOCKET_OPERATION_PONG = 10

WEBSOCKET_DATA_OPERATIONS = (WEBSOCKET_OPERATION_CONTINUOUS,  WEBSOCKET_OPERATION_TEXT, WEBSOCKET_OPERATION_BINARY)
WEBSOCKET_CONTROL_OPERATIONS = (WEBSOCKET_OPERATION_CLOSE, WEBSOCKET_OPERATION_PING, WEBSOCKET_OPERATION_PONG)

# TODO: whats the fastest way on pypy ? casting 64 bit ints -> xor -> replace?
_XOR_TABLE = [bytes(a^b for a in range(256)) for b in range(256)]

def apply_websocket_mask(mask, data):
    """
    Applies websocket mask on the given data and returns a new instance.
    
    Parameters
    ----------
    data : `bytes-like`
        Data to apply the mask to.
    mask : `bytes`
        `uint32` websocket mask in bytes.
    """
    data_bytes = bytearray(data)
    for index in range(4):
        data_bytes[index::4] = data_bytes[index::4].translate(_XOR_TABLE[mask[index]])
    return data_bytes


class WebSocketFrame:
    """
    Represents a websocket frame.
    
    Attributes
    ----------
    data : `bytes`
        The data of the frame.
    head_1 : `int`
        The first bytes of websocket frame's, because it holds all the required data needed.
    """
    __slots__ = ('data', 'head_1',)
    
    def __init__(self, is_final, operation_code, data):
        """
        Creates a websocket frame.
        
        Parameters
        -----------
        is_final : `bool`
            Whether this websocket frame is a final websocket frame. When sending websocket frames, the data of frames
            it collected, till a final frame is received.
        operation_code : `int`
            The operation code of the websocket frame.
            
            Can be 1 of the following:
            
            +-----------------------------------+-------+
            | Respective name                   | Value |
            +===================================+=======+
            | WEBSOCKET_OPERATION_CONTINUOUS    | 0     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_TEXT          | 1     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_BINARY        | 2     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_CLOSE         | 8     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_PING          | 9     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_PONG          | 10    |
            +-----------------------------------+-------+
        
        data : `bytes`
            The data to ship with the websocket frame.
        """
        self.data = data
        self.head_1 = (is_final << 7) | operation_code
    
    @property
    def is_final(self):
        """
        Returns whether the websocket frame is final.
        
        Returns
        -------
        is_final : `bool`
        """
        return (self.head_1 & 0b10000000) >> 7
    
    
    @property
    def rsv1(self):
        """
        Returns the first reserved bit of the websocket frame's head.
        
        Defaults to `0˙
        
        Returns
        -------
        rsv1 : `int`
        """
        return (self.head_1 & 0b01000000) >> 6
    
    
    @property
    def rsv2(self):
        """
        Returns the second reserved bit of the websocket frame's head.
        
        Defaults to `0˙
        
        Returns
        -------
        rsv2 : `int`
        """
        return (self.head_1 & 0b00100000) >> 5
    
    
    @property
    def rsv3(self):
        """
        Returns the third reserved bit of the websocket frame's head.
        
        Defaults to `0˙
        
        Returns
        -------
        rsv3 : `int`
        """
        return (self.head_1 & 0b00010000) >> 4
    
    
    @property
    def operation_code(self):
        """
        Returns the websocket frame's operation_code.
        
        Returns
        -------
        operation_code : `int`
            Can be one of the following values:
            
            +-----------------------------------+-------+
            | Respective name                   | Value |
            +===================================+=======+
            | WEBSOCKET_OPERATION_CONTINUOUS    | 0     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_TEXT          | 1     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_BINARY        | 2     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_CLOSE         | 8     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_PING          | 9     |
            +-----------------------------------+-------+
            | WEBSOCKET_OPERATION_PONG          | 10    |
            +-----------------------------------+-------+
        """
        return  self.head_1 & 0b00001111
    
    
    def check(self):
        """
        Check that this frame contains acceptable values.
        
        Raises
        ------
        WebSocketProtocolError
            - If the reserved bits are not `0`.
            - If the frame is a control frame, but is too long for one.
            - If the websocket frame is fragmented frame. (Might be supported if people request is.)
            - If the frame operation_code is not any of the expected ones.
        """
        if self.head_1 & 0b01110000:
            raise WebSocketProtocolError('Reserved bits must be `0`.')
        
        operation_code = self.head_1 & 0b00001111
        if operation_code in WEBSOCKET_DATA_OPERATIONS:
            return
        
        if operation_code in WEBSOCKET_CONTROL_OPERATIONS:
            if len(self.data) > 125:
                raise WebSocketProtocolError('Control frame too long.')
            if not self.head_1 & 0b10000000:
                raise WebSocketProtocolError('Fragmented control frame.')
            return
        
        raise WebSocketProtocolError(f'Invalid operation_code: {operation_code}.')
    
    
    @classmethod
    def _from_fields(cls, data, head_1):
        """
        Creates a new websocket frame instance.
        
        Parameters
        ----------
        data : `bytes`
            The data of the frame.
        head_1 : `int`
            The first bytes of websocket frame's, because it holds all the required data needed.
        
        Returns
        -------
        self : ``WebSocketFrame``
        """
        self = object.__new__(cls)
        self.data = data
        self.head_1 = head_1
        return self



__all__ = ('alchemy_incendiary', 'call', 'copy_func',)

from types import FunctionType

from .docs import has_docs
from .trace import ignore_frame


ignore_frame(__spec__.origin, '__call__', 'return self.func(*self.args)', )
ignore_frame(__spec__.origin, '__call__', 'return self.func(*self.args, **kwargs)', )

@has_docs
def call(function):
    """
    Calls the function returning itself.
    
    Parameters
    ----------
    function : `Any`
        The function to call
    
    Returns
    -------
    function : `Any`
        The function itself.
    """
    function()
    return function


@has_docs
class alchemy_incendiary:
    """
    Function wrapper familiar to `functools.partial`.
    
    Used by hata to run functions inside of executors.
    
    Attributes
    ----------
    args : `tuple` of `Any`
        Parameters to call `func` with.
    func : `callable`
        The function to call.
    kwargs : `None` of `dict` of (`str`, `Any`) items
        Keyword parameters to call func with if applicable.
    """
    __slots__ = ('args', 'func', 'kwargs',)
    
    @has_docs
    def __init__(self, func, args, kwargs=None):
        """
        Creates a new `alchemy_incendiary` with the given parameters.
        
        Parameters
        ----------
        func : `callable`
            The function to call.
        args : `tuple` of `Any`
            Parameters to call `func` with.
        kwargs : `None` of `dict` of (`str`, `Any`) items = `None`, Optional
            Keyword parameters to call func with if applicable.
        """
        self.func = func
        self.args = args
        self.kwargs = kwargs
    
    @has_docs
    def __call__(self):
        """
        Calls the ``alchemy_incendiary``'s inner function with t's parameters and keyword parameters.
        
        Returns
        -------
        result : `Any`
            The returned value by ``.func``.
        
        Raises
        ------
        BaseException
            The raised exception by ``.func``.
        """
        kwargs = self.kwargs
        if kwargs is None:
            return self.func(*self.args)
        
        return self.func(*self.args, **kwargs)


@has_docs
def copy_func(old):
    """
    Copies the given function.
    
    Parameters
    ----------
    old : `function`
        The function to copy.
    
    Returns
    -------
    new : `functions`
        The new created function.
    """
    new = FunctionType(old.__code__, old.__globals__, name=old.__name__, argdefs=old.__defaults__,
        closure=old.__closure__)
    new.__kwdefaults__ = old.__kwdefaults__
    return new



__all__ = (
    'BROTLI_COMPRESSOR', 'BROTLI_DECOMPRESSOR', 'COMPRESSION_ERRORS', 'ZLIB_DECOMPRESSOR', 'ZLIB_COMPRESSOR',
    'ZLIB_MAX_WBITS', 'get_decompressor_for'
)

import zlib

from .exceptions import ContentEncodingError


try:
    import brotli
except ImportError:
    brotli = None
    COMPRESSION_ERRORS = zlib.error
    BROTLI_DECOMPRESSOR = None
    BROTLI_COMPRESSOR = None
else:
    COMPRESSION_ERRORS = (zlib.error, brotli.error)
    
    if hasattr(brotli, 'Error'):
        # brotlipy case
        BROTLI_DECOMPRESSOR = brotli.Decompressor
        BROTLI_COMPRESSOR = brotli.Compressor
    else:
        # brotli case
        class BROTLI_DECOMPRESSOR:
            __slots__ = ('_decompressor', )
            def __init__(self):
                self._decompressor = brotli.Decompressor()
            
            def decompress(self, value):
                return self._decompressor.process(value)
        
        class BROTLI_COMPRESSOR:
            __slots__ = ('_compressor', )
            def __init__(self):
                self._compressor = brotli.Compressor()
            
            def compress(self, value):
                return self._compressor.process(value)


ZLIB_DECOMPRESSOR = zlib.decompressobj
ZLIB_COMPRESSOR = zlib.compressobj

ZLIB_MAX_WBITS = zlib.MAX_WBITS

def get_decompressor_for(content_encoding):
    """
    Gets decompress object for the given content-encoding if applicable.
    
    Parameters
    ----------
    content_encoding : `None`, `str`
        Content encoding of a respective http message.
    
    Returns
    -------
    decompressor : `None`, `ZLIB_DECOMPRESSOR`, `BROTLI_DECOMPRESSOR`
    
    Raises
    ------
    ContentEncodingError
        - `'content_encoding'` was given as `'br'` meanwhile brotli or brotlipy are not installed.
        - `'content_encoding'` is not an from the expected values.
    """
    if (content_encoding is None):
        decompressor = None
    
    elif content_encoding == 'gzip':
        decompressor = ZLIB_DECOMPRESSOR(wbits=16 + ZLIB_MAX_WBITS)
    
    elif content_encoding == 'deflate':
        decompressor = ZLIB_DECOMPRESSOR(wbits=-ZLIB_MAX_WBITS)
    
    elif content_encoding == 'br':
        if BROTLI_DECOMPRESSOR is None:
            raise ContentEncodingError(
                'Can not decode content-encoding: brotli (br). Please install `brotlipy`.'
            )
        decompressor = BROTLI_DECOMPRESSOR()
    
    elif content_encoding == 'identity':
        # I assume this is no encoding
        decompressor = None
    
    else:
        raise ContentEncodingError(
            f'Can not decode content-encoding: {content_encoding!r}.'
        )
    
    return decompressor



__all__ = ()


class WordNode:
    """
    A words' character node when building regex.
    
    Attributes
    ----------
    character : `str`
        The represented character by the node.
    is_final : `bool`
        Whether the node is the end of a word.
    nodes : `dict` of (`str`, ``WordNode``) items
        Sub nodes branching out.
    parent : `None`, ``WordNode``
        The parent node.
    """
    __slots__ = ('character', 'is_final', 'nodes', 'parent')
    
    def __new__(cls, character, is_final, parent):
        """
        Creates a new ``WordNode`` with the given `character`.
        
        Parameters
        ----------
        character : `str`
            The character of the node.
        is_final : `bool`
            Whether the node is the end of a word.
        """
        self = object.__new__(cls)
        self.character = character
        self.nodes = None
        self.is_final = is_final
        self.parent = parent
        return self
    
    
    def __repr__(self):
        """Returns the word node's representation."""
        result = ['<', self.__class__.__name__, ' of ', repr(self.character)]
        if self.is_final:
            result.append(' (final)')
        
        nodes = self.nodes
        if (nodes is not None):
            result.append(' nodes=')
            result.append(repr(nodes))
        
        result.append('>')
        
        return ''.join(result)
    
    
    def add_node(self, characters, character_index):
        """
        Adds a sub-node to the node.
        
        Attributes
        ----------
        characters : `list` of `str`
            A list of characters to add as nodes.
        character_index : `int`
            The character's index to use from `characters`.
        """
        character = characters[character_index]
        character_index += 1
        if len(characters) == character_index:
            is_final = True
        else:
            is_final = False
        
        nodes = self.nodes
        if (nodes is None):
            self.nodes = nodes = {}
        
        try:
            node = nodes[character]
        except KeyError:
            node = WordNode(character, is_final, self)
            self.nodes[character] = node
        else:
            if is_final:
                node.is_final = True
        
        if (not is_final):
            node.add_node(characters, character_index)
    
    
    def _match_index(self, string, index):
        """
        Matches the pattern from the given string.
        
        Parameters
        ----------
        string : `str`
            The string to match from.
        index : `int`
            The starter index to match since. Defaults to `0`.
        
        Returns
        -------
        matched_count : `int`
            The amount of matched characters.
            
            `-1` is returned of nothing is matched.
        """
        if len(string) > index:
            character = string[index]
            
            nodes = self.nodes
            if (nodes is not None):
                try:
                    node = nodes[character]
                except KeyError:
                    pass
                else:
                    match_index = node._match_index(string, index + 1)
                    if match_index != -1:
                        return match_index + 1
        
        if self.is_final:
            return 1
        
        return -1
    
    def match(self, string, index=0):
        """
        Matches pattern from the given string.
        
        Parameters
        ----------
        string : `str`
            The string to match from.
        index : `int` = `0`, Optional
            The starter index to match since. Defaults to `0`.
        
        Returns
        -------
        matched : `None`, `str`
            The matched string, if any.
        """
        if len(string) > index:
            character = string[index]
            
            nodes = self.nodes
            if (nodes is not None):
                try:
                    node = nodes[character]
                except KeyError:
                    pass
                else:
                    match_index = node._match_index(string, index + 1)
                    if match_index != -1:
                        return string[index:index + match_index]
        
        if self.is_final:
            return ''
        
        return None


def create_word_pattern(words):
    """
    Creates 1 regex pattern from many words.
    
    Parameters
    ----------
    words : `iterable` of `str`
        The words to create regex pattern from.
    
    Returns
    -------
    regex_pattern : ``WordNode``
        The generated pattern.
    """
    word_node = WordNode('', False, None)
    for word in words:
        if word:
            word_node.add_node(list(word), 0)
        else:
            word_node.is_final = True
    
    return word_node



__all__ = (
    'AbortHandshake', 'ConnectionClosed', 'ContentEncodingError', 'HttpProcessingError', 'InvalidHandshake',
    'InvalidOrigin', 'InvalidUpgrade', 'PayloadError', 'ProxyError', 'WebSocketProtocolError'
)

class PayloadError(Exception):
    """
    Raised when http payload processing fails.
    """
    pass


class InvalidHandshake(Exception):
    """
    Raised when websocket handshake fails.
    
    Attributes
    ----------
    message : `str`
        Error message.
    response : `None`, ``ClientResponse``
        Received http answer.
    request : ``RawRequestMessage``
        Received raw http request.
    """
    def __init__(self, message, *, response=None, request=None):
        self.response = response
        self.message = message
        self.request = request
        Exception.__init__(self, message)


class HttpProcessingError(Exception):
    """
    Base class for http content specific errors.
    
    Attributes
    ----------
    code : `int`
        Http error code. Defaults to `0`.
    message : `str`
        Error message. Defaults to empty string.
    headers : `None`, ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Respective headers.
    """
    def __init__(self, message='', code=0, headers=None):
        self.code = code
        self.headers = headers
        self.message = message
        
        Exception.__init__(self, f'HTTP {code}, message={message!r}, headers={headers!r}')


class AbortHandshake(HttpProcessingError, InvalidHandshake):
    """
    Raised when websocket handshake is aborted on server side.
    
    Attributes
    ----------
    code : `int`
        Http error code. Defaults to `0`.
    message : `str`
        Error message. Defaults to empty string.
    headers : `None`, ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Respective headers.
    response : `None`, ``ClientResponse``
        Received http answer.
    request : ``RawRequestMessage``
        Received raw http request.
    """
    def __init__(self, message='', code=0, headers=None, *, response=None, request=None):
        self.response = response
        self.message = message
        self.request = request
        self.code = code
        self.headers = headers
        
        Exception.__init__(self, f'HTTP {code}, message={message!r}, headers={headers!r}')


class ProxyError(HttpProcessingError):
    """
    Raised when a proxy request responds with status other than "200 OK".
    
    Attributes
    ----------
    code : `int`
        Http error code. Defaults to `0`.
    message : `str`
        Error message. Defaults to empty string.
    headers : `None`, ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Respective headers.
    """
    pass


class InvalidOrigin(InvalidHandshake):
    """
    Raised when a websocket handshake received invalid origin header.
    """
    pass


class InvalidUpgrade(InvalidHandshake):
    """
    Raised when a websocket was not correctly upgraded.
    """
    pass


class ContentEncodingError(HttpProcessingError, PayloadError):
    """
    Raised when http content decoding fails.
    
    Attributes
    ----------
    code : `int`
        Http error code. Defaults to `0`.
    message : `str`
        Error message. Defaults to empty string.
    headers : `None`, ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Respective headers.
    """
    def __init__(self, message='Bad Request', headers=None):
        HttpProcessingError.__init__(self, message, 400, headers)


CLOSE_REASONS = {
    1000: 'OK',
    1001: 'going away',
    1002: 'protocol error',
    1003: 'unsupported type',
    1004: '`reserved`',
    1005: 'no status code [internal]',
    1006: 'connection closed abnormally [internal]',
    1007: 'invalid data',
    1008: 'policy violation',
    1009: 'message too big',
    1010: 'extension required',
    1011: 'unexpected error',
    1013: 'Try again later',
    1014: 'Bad gateway',
    1015: 'TLS failure [internal]',
}

def get_close_reason(code):
    """
    Gets close for any websocket close code.
    
    Parameters
    ----------
    code : `int`
        Web socket close code.
    
    Returns
    -------
    reason : `str`
    """
    try:
        close_reason = CLOSE_REASONS[code]
    except KeyError:
        if code < 1000:
            close_reason = '`unused`'
        elif code < 2000:
            close_reason = '`reserved`'
        elif code < 3000:
            close_reason = '`reserved for extensions`'
        elif code < 4000:
            close_reason = '`registered`'
        elif code < 5000:
            close_reason = '`private use`'
        else:
            close_reason = '`unknown`'
    
    return close_reason

class ConnectionClosed(Exception):
    """
    Connection closed exception raised when a websocket is closed.
    
    Attributes
    ----------
    code : `int`
        Web socket close code.
    exception : `None`, `BaseException`
        Source exception if applicable.
    reason : `None or `str`
        Web socket close reason if any.
    """
    
    def __init__(self, code, exception, reason=None):
        """
        Creates a new ``ConnectionClosed`` exception from the given parameters.
        
        Parameters
        ----------
        code : `int`
            The websocket close code.
        exception : `None`, `BaseException`
            Source exception if applicable.
        reason : `None`, `str` = `None`, Optional
            Web socket close reason if any.
        """
        self.code = code
        self.exception = exception
        self._reason = reason
        Exception.__init__(self)
    
    @property
    def reason(self):
        """
        Returns the websocket close reason.
        
        Returns
        -------
        reason : `str`
        """
        reason = self._reason
        if (reason is None):
            reason = get_close_reason(self.code)
        
        return reason
    
    def __repr__(self):
        """Returns the exception's representation."""
        return f'<{self.__class__.__name__}, code={self.code!r}, reason={self.reason!r}, exception={self.exception!r}>'
    
    __str__ = __repr__


class WebSocketProtocolError(Exception):
    """
    Exception raised by websocket when receiving invalid payload.
    """
    pass



__all__ = ('ClientRequest',)

from http.cookies import Morsel, SimpleCookie

from ..core import CancelledError, Task
from ..utils import IgnoreCaseMultiValueDictionary
from ..web_common.formdata import Formdata
from ..web_common.headers import (
    ACCEPT, ACCEPT_ENCODING, AUTHORIZATION, CONTENT_ENCODING, CONTENT_LENGTH, CONTENT_TYPE, COOKIE, HOST,
    METHOD_CONNECT, METHOD_POST_ALL, TRANSFER_ENCODING
)
from ..web_common.helpers import BasicAuth
from ..web_common.http_stream_writer import HTTPStreamWriter
from ..web_common.multipart import create_payload

from .client_response import ClientResponse
from .connection_key import ConnectionKey
from .request_info import RequestInfo


DEFAULT_HEADERS = (
    (ACCEPT, '*/*'),
    (ACCEPT_ENCODING, 'gzip, deflate'),
)

class ClientRequest:
    """
    Http request class used by ``HTTPClient``.
    
    Attributes
    ----------
    auth : `None`, ``BasicAuth``
        Authorization sent with the request.
    body : `None`, ``PayloadBase``
        The request's body.
    chunked : `bool`
        Whether the request is sent chunked.
    compression : `None`, `str`
        Compression used when sending the request.
    headers : `IgnoreCaseMultiValueDictionary`
        The headers of the request.
    loop : ``EventThread``
        The event loop, trough what the request is executed.
    method : `str`
        The request's method.
    original_url : ``URL``
        The original url, what was asked to request.
    proxy_auth : `None`, ``BasicAuth``
        Proxy authorization sent with the request.
    proxy_url : `None`, ``URL``
        Proxy url to use if applicable.
    response : `None`, ``ClientResponse``
        Object representing the received response. Set as `None` till ``.send`` finishes.
    ssl : `None` `None`, ``SSLContext``, `bool`, ``Fingerprint``
        The connection's ssl type.
    url : ``URL``
        The url, what will be requested.
    writer : `None`, ``Task`` of ``.write_bytes``
        Payload writer task, what is present meanwhile the request's payload is sending.
    """
    __slots__ = (
        'auth', 'body', 'chunked', 'compression', 'headers', 'loop', 'method', 'original_url', 'proxy_auth',
        'proxy_url', 'response', 'ssl', 'url', 'writer'
    )
    
    def __new__(cls, loop, method, url, headers, data, params, cookies, auth, proxy_url, proxy_auth, ssl):
        """
        Creates a new ``ClientRequest`` with the given parameters.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop, trough what the request is executed.
        method : `str`
            The request's method.
        url : ``URL``
            The url to request.
        headers : `None`, `dict`, ``IgnoreCaseMultiValueDictionary``
            Headers of the request.
        data : `None`, `bytes-like`, `io-like`, ``Formdata`
            Data to send as the request's body.
        params : `dict` of (`str`, (`str`, `int`, `float`, `bool`)) items
            Query string parameters.
        cookies : `None`, ``CookieJar``
            Cookies OwO.
        auth : `None`, ``BasicAuth``
            Authorization sent with the request.
        proxy_url : `None`, ``URL``
            Proxy url to use if applicable.
        proxy_auth : `None`, ``BasicAuth``
            Proxy authorization sent with the request.
        ssl : `None` `None`, ``SSLContext``, `bool`, ``Fingerprint``
            The connection's ssl type.
        
        Raises
        ------
        TypeError
            - `proxy_auth`'s type is incorrect.
            - ˙Cannot serialize a field of the given `data`.
        ValueError
            - Host could not be detected from `url`.
            - The `proxy_url`'s scheme is not `http`.
            - `compression` and `Content-Encoding` would be set at the same time.
            - `chunked` cannot be set, because `Transfer-Encoding: chunked` is already set.
            - `chunked` cannot be set, because `Content-Length` header is already present.
        RuntimeError
            - If one of `data`'s field's content has unknown content-encoding.
            - If one of `data`'s field's content has unknown content-transfer-encoding.
        """
        # Convert headers
        headers = IgnoreCaseMultiValueDictionary(headers)
        
        # Add extra query parameters to the url and remove fragments
        url = url.extend_query(params)
        request_url = url.with_fragment(None)
        
        if not url.host:
            raise ValueError('Host could not be detected.')
        
        # Check authorization
        if auth is None:
            # If authorization is given, try to detect from url.
            username = url.user
            password = url.password
            
            if (username is not None) and username:
                if password is None:
                    password = ''
                
                auth = BasicAuth(username, password)
        
        # Store auth in headers is applicable.
        if (auth is not None):
            headers[AUTHORIZATION] = auth.encode()
        
        for key, value in DEFAULT_HEADERS:
            headers.setdefault(key, value)
        
        # Add host to headers if not present.
        if HOST not in headers:
            netloc = request_url.raw_host
            if not request_url.is_default_port():
                netloc = f'{netloc}:{request_url.port}'
            
            headers[HOST] = netloc
        
        # Update cookies
        if (cookies is not None) and cookies:
            cookie = SimpleCookie()
            if COOKIE in headers:
                cookie.load(headers.get(COOKIE, ''))
                del headers[COOKIE]
            
            for key, value in cookies.items():
                if isinstance(key, Morsel):
                    # Preserve coded_value
                    try:
                        morsel_value = value.get(value.key, None)
                    except KeyError:
                        morsel_value = Morsel()
                    
                    morsel_value.set(value.key, value.value, value.coded_value)
                    value = morsel_value
                
                cookie[key] = value
            
            headers[COOKIE] = cookie.output(header='', sep=';').strip()
        
        # Check proxy settings.
        if proxy_url is not None:
            if proxy_url.scheme != 'http':
                raise ValueError(
                    f'Only http proxies are supported, got {proxy_url!r}.'
                )
            
            if (proxy_auth is not None) and not isinstance(proxy_auth, BasicAuth):
                raise TypeError(
                    f'`proxy_auth` can be `None`, `{BasicAuth.__name__}`, got '
                    f'{proxy_auth.__class__.__name__}; {proxy_auth!r}.'
                )
        
        # Needed for transfer data checks
        chunked = True
        compression = headers.get(CONTENT_ENCODING, None)
        
        # Get request content encoding.
        if (data is not None):
            if data:
                if (compression is not None):
                    if headers.get(CONTENT_ENCODING, ''):
                        raise ValueError(
                            f'Compression can not be set if `Content-Encoding` header is set.'
                        )
                    
                    chunked = True
                
                # formdata
                if isinstance(data, Formdata):
                    data = data()
                else:
                    try:
                        data = create_payload(data, {'disposition': None})
                    except LookupError:
                        data = Formdata.from_fields(data)()
                
                if not chunked:
                    if CONTENT_LENGTH not in headers:
                        size = data.size
                        if size is None:
                            chunked = True
                        else:
                            if CONTENT_LENGTH not in headers:
                                headers[CONTENT_LENGTH] = str(size)
                
                if CONTENT_TYPE not in headers:
                    headers[CONTENT_TYPE] = data.content_type
                
                data_headers = data.headers
                if data_headers:
                    for key, value in data_headers.items():
                        headers.setdefault(key, value)
            else:
                data = None
        
        # Analyze transfer-encoding header.
        transfer_encoding = headers.get(TRANSFER_ENCODING, '').lower()
        
        if 'chunked' in transfer_encoding:
            if chunked:
                raise ValueError(
                    f'Chunked can not be set if `Transfer-Encoding: chunked` header is already set.'
                )
        
        elif chunked:
            if CONTENT_LENGTH in headers:
                raise ValueError(
                    'Chunked can not be set if `Content-Length` header is set.'
                )
            headers[TRANSFER_ENCODING] = 'chunked'
        
        else:
            if CONTENT_LENGTH not in headers:
                headers[CONTENT_LENGTH] = '0' if data is None else str(len(data))
        
        # Set default content-type.
        if (method in METHOD_POST_ALL) and (CONTENT_TYPE not in headers):
            headers[CONTENT_TYPE] = 'application/octet-stream'
        
        # Everything seems correct, create the object.
        self = object.__new__(cls)
        self.original_url = url
        self.url = request_url
        self.method = method
        self.loop = loop
        self.ssl = ssl
        self.chunked = chunked
        self.compression = compression
        self.body = data
        self.auth = auth
        self.writer = None
        self.response = None
        self.headers = headers
        self.proxy_url = proxy_url
        self.proxy_auth = proxy_auth
        
        return self
    
    def is_ssl(self):
        """
        Returns whether the request is ssl.
        
        Returns
        -------
        is_ssl : `bool`
        """
        return self.url.scheme in ('https', 'wss')
    
    @property
    def connection_key(self):
        """
        Returns the connection key of request.
        
        Returns
        -------
        connection_key : ``ConnectionKey``
        """
        return ConnectionKey(self)
    
    @property
    def request_info(self):
        """
        Returns base information representing the request.
        
        Returns
        -------
        request_info : ``RequestInfo``
        """
        return RequestInfo(self)
    
    @property
    def host(self):
        """
        Returns the request's host.
        
        Returns
        -------
        host : `str`
        """
        return self.url.host
    
    @property
    def port(self):
        """
        Returns the request's port.
        
        Returns
        -------
        port : `int`
        """
        return self.url.port
    
    async def write_bytes(self, writer, connection):
        """
        Writes the request's body..
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Writer used to write the request's body into the connection's transport.
        connection : ``Connection``
            Connection of the request with what the payload is sent.
        """
        # Support coroutines that yields bytes objects.
        try:
            body = self.body
            if (body is not None):
                await self.body.write(writer)
            await writer.write_eof()
        except OSError as err:
            new_err = OSError(err.errno, f'Can not write request body for {self.url!r}.')
            new_err.__context__ = err
            new_err.__cause__ = err
            connection.protocol.set_exception(new_err)
        
        except CancelledError as err:
            if not connection.closed:
               connection.protocol.set_exception(err)
        
        except BaseException as err:
            connection.protocol.set_exception(err)
            raise
        
        finally:
            self.writer = None
    
    
    def send(self, connection):
        """
        Sends the request.
        
        Parameters
        ----------
        connection : ``Connection``
            Connection, what is used to send the request.
        
        Returns
        -------
        response : `CoroutineType` of ``ClientResponse.start`` ->
        """
        try:
            url = self.url
            if self.method == METHOD_CONNECT:
                path = f'{url.raw_host}:{url.port}'
            elif (self.proxy_url is not None) and (not self.is_ssl()):
                path = str(url)
            else:
                path = url.raw_path
                if url.raw_query_string:
                    path = f'{path}?{url.raw_query_string}'
            
            protocol = connection.protocol
            writer = HTTPStreamWriter(protocol, self.compression, self.chunked)
            
            protocol.write_http_request(self.method, path, self.headers)
            
            self.writer = Task(self.write_bytes(writer, connection), self.loop)
            
            self.response = response = ClientResponse(self, connection)
            
            return response.start()
        
        except:
            connection.close()
            raise
    
    
    def terminate(self):
        """
        Terminates the request's writing task if applicable.
        """
        writer = self.writer
        if (writer is not None):
            self.writer = None
            writer.cancel()
